from dataclasses import dataclass, field
from typing import Any, Literal, Optional

from google.genai.types import Part
from matrx_utils import vcprint
from openai.types.responses import (
    ResponseOutputText as OpenAIResponseOutputText,
)
from openai.types.responses import (
    ResponseReasoningItem as OpenAIResponseReasoningItem,
)

from matrx_ai.config.extra_config import (
    CodeExecutionContent,
    CodeExecutionResultContent,
    WebSearchCallContent,
)
from matrx_ai.config.media_config import (
    AudioContent,
    DocumentContent,
    ImageContent,
    VideoContent,
    YouTubeVideoContent,
    reconstruct_media_content,
)

from .tools_config import ToolCallContent, ToolResultContent

# ============================================================================
# CORE UNIFIED TYPES (Provider-Agnostic - Pure/Reusable)
# These types are completely independent of AI Matrix
# ============================================================================


@dataclass
class TextContent:
    type: Literal["text"] = "text"
    text: str = ""
    id: str = ""
    metadata: dict[str, Any] = field(default_factory=dict)

    def get_output(self) -> str:
        """Get the output text of the content."""
        return self.text

    def replace_variables(self, variables: dict[str, Any]) -> None:
        """
        Replace variables in text content.
        Variables are in format {{variable_name}} in the text.

        Args:
            variables: dict mapping variable names to their values
        """
        for var_name, var_value in variables.items():
            self.text = self.text.replace(f"{{{{{var_name}}}}}", str(var_value))

    def append_text(self, text: str, separator: str = "\n") -> None:
        """
        Append text to existing content.

        Args:
            text: The text to append
            separator: Separator between existing and new text (default: newline)
        """
        self.text += f"{separator}{text}"

    def to_google(self) -> dict[str, Any]:
        """Convert to Google Gemini format"""
        part = {"text": self.text}
        # Retrieve Google's thought signature from metadata if present
        if "google_thought_signature" in self.metadata:
            part["thoughtSignature"] = self.metadata["google_thought_signature"]
        return part

    def to_openai(self, role: str | None = None) -> dict[str, Any]:
        """Convert to OpenAI format"""
        # Assistant messages use output_text, all others use input_text
        text_type = "output_text" if role == "assistant" else "input_text"
        return {"type": text_type, "text": self.text}

    def to_anthropic(self) -> dict[str, Any]:
        """Convert to Anthropic format"""
        return {"type": "text", "text": self.text}

    def _sanitize_metadata(self, metadata: dict[str, Any]) -> dict[str, Any]:
        """
        Sanitize metadata by replacing sensitive encrypted content with length information.

        Replaces these fields with their lengths:
        - anthropic_signature: str
        - google_thought_signature: bytes
        - encrypted_content: bytes
        """
        sanitized = metadata.copy()

        # Replace sensitive fields with length info
        if "anthropic_signature" in sanitized:
            sanitized["anthropic_signature"] = (
                f"<str length={len(sanitized['anthropic_signature'])}>"
            )

        if "google_thought_signature" in sanitized:
            sanitized["google_thought_signature"] = (
                f"<bytes length={len(sanitized['google_thought_signature'])}>"
            )

        if "encrypted_content" in sanitized:
            sanitized["encrypted_content"] = (
                f"<bytes length={len(sanitized['encrypted_content'])}>"
            )

        return sanitized

    def to_dict(self) -> dict[str, Any]:
        """
        Convert to dictionary with sanitized metadata.
        Use this instead of dataclasses.asdict() to avoid printing sensitive encrypted content.
        """
        return {
            "type": self.type,
            "text": self.text,
            "id": self.id,
            "metadata": self._sanitize_metadata(self.metadata),
        }

    def to_storage_dict(self) -> dict[str, Any]:
        """Serialize to storage format for database persistence (cx_message.content JSONB)."""
        result: dict[str, Any] = {"type": "text", "text": self.text}
        if self.id:
            result["id"] = self.id
        citations = self.metadata.get("citations")
        if citations:
            result["citations"] = citations
        return result

    def __repr__(self) -> str:
        """Override repr to show sanitized metadata instead of full encrypted content"""
        sanitized_metadata = self._sanitize_metadata(self.metadata)

        # Show first 35 and last 35 characters with ... in the middle
        if len(self.text) > 70:
            text_preview = f"{self.text[:35]}...{self.text[-35:]}"
        else:
            text_preview = self.text

        return (
            f"TextContent(type={self.type!r}, "
            f"text={text_preview!r}, "
            f"id={self.id!r}, "
            f"metadata={sanitized_metadata!r})"
        )

    @classmethod
    def from_openai(
        cls, content_item: OpenAIResponseOutputText, id: str
    ) -> Optional["TextContent"]:
        text = content_item.text
        metadata = content_item.model_dump(exclude={"text"})
        metadata["id"] = id

        return cls(
            id=id,
            text=text,
            metadata=metadata,
        )

    @classmethod
    def from_google(cls, part: Part) -> Optional["TextContent"]:
        """Create TextContent from Google Part object"""
        metadata = {}
        # Store Google's thought signature in metadata if present
        if part.thought_signature:
            metadata["google_thought_signature"] = part.thought_signature
        return cls(
            text=part.text,
            metadata=metadata,
        )

    @classmethod
    def from_anthropic(cls, content_block: dict[str, Any]) -> Optional["TextContent"]:
        """Create TextContent from Anthropic content block"""
        metadata = {}
        citations = content_block.get("citations", [])
        if citations:
            metadata["citations"] = citations
        return cls(
            text=content_block["text"],
            metadata=metadata,
        )


@dataclass
class ThinkingContent:
    """
    Extended thinking content with normalized provider identification.

    Normalization:
    - provider: Explicitly identifies which AI provider generated this content
    - signature: Unified field for provider-specific encrypted/signature data
      (OpenAI's encrypted_content, Anthropic's signature, Google's thought_signature)

    Each provider can only process its own signature data. The to_* methods
    check the provider field and only include signature if it matches.
    """

    type: Literal["thinking"] = "thinking"
    text: str = ""
    id: str = ""
    summary: list[dict[str, Any]] = field(default_factory=list)

    # Normalized fields for database storage
    provider: Literal["openai", "anthropic", "google", "cerebras"] | None = None
    signature: str | bytes | None = (
        None  # Provider-specific encrypted/signature data
    )

    # metadata for truly optional/non-essential data only
    metadata: dict[str, Any] = field(default_factory=dict)

    def get_output(self) -> str:
        text_output = ""
        if self.text:
            text_output += self.text
        if self.summary:
            text_output += "\n".join([item["text"] for item in self.summary])
        return text_output

    def _sanitize_signature(self) -> str:
        """Return a sanitized representation of the signature for display."""
        if self.signature is None:
            return "None"
        if isinstance(self.signature, bytes):
            return f"<bytes length={len(self.signature)}>"
        return f"<str length={len(self.signature)}>"

    def to_dict(self) -> dict[str, Any]:
        """
        Convert to dictionary with sanitized signature.
        Use this instead of dataclasses.asdict() to avoid printing sensitive encrypted content.
        """
        result = {
            "type": self.type,
            "text": self.text,
            "id": self.id,
            "summary": self.summary,
            "provider": self.provider,
            "signature": self.signature,
        }
        if self.metadata:
            result["metadata"] = self.metadata
        return result

    def to_storage_dict(self) -> dict[str, Any]:
        """Serialize to storage format for database persistence (cx_message.content JSONB)."""
        import base64

        result: dict[str, Any] = {
            "type": "thinking",
            "text": self.text,
            "provider": self.provider,
            "metadata": self.metadata,
        }

        if self.signature:
            # Signature can be bytes (from providers) — base64-encode for JSONB storage
            if isinstance(self.signature, bytes):
                result["signature"] = base64.b64encode(self.signature).decode("ascii")
                result["signature_encoding"] = "base64"
            else:
                result["signature"] = self.signature

        if self.summary:
            # summary items may be Pydantic models (e.g. OpenAI Summary) or plain dicts
            result["summary"] = [
                item.model_dump() if hasattr(item, "model_dump") else item
                for item in self.summary
            ]

        if self.id:
            result["metadata"]["id"] = self.id

        return result

    def __repr__(self) -> str:
        """Override repr to show sanitized signature instead of full encrypted content"""
        # Show first 35 and last 35 characters with ... in the middle
        if len(self.text) > 70:
            text_preview = f"{self.text[:35]}...{self.text[-35:]}"
        else:
            text_preview = self.text

        return (
            f"ThinkingContent(type={self.type!r}, "
            f"text={text_preview!r}, "
            f"id={self.id!r}, "
            f"summary={self.summary!r}, "
            f"provider={self.provider!r}, "
            f"signature={self._sanitize_signature()})"
        )

    def to_google(self) -> dict[str, Any] | None:
        """Convert to Google Gemini format. Only includes signature if provider is Google."""
        # Only include signature if this content came from Google
        if self.provider != "google" or self.signature is None:
            return None

        part = {"text": self.text, "thought": True}
        part["thoughtSignature"] = self.signature
        return part

    def to_openai(self) -> dict[str, Any] | None:
        """Convert to OpenAI format. Only includes signature if provider is OpenAI."""
        # Only include encrypted_content if this content came from OpenAI
        if self.provider != "openai" or self.signature is None:
            return None

        return {
            "id": self.id,
            "summary": self.summary,
            "type": "reasoning",
            "encrypted_content": self.signature,
        }

    def to_anthropic(self) -> dict[str, Any] | None:
        """
        Convert to Anthropic format. Only includes signature if provider is Anthropic.
        Returns None if signature is missing (Anthropic requires signature for thinking).
        """
        # Only include signature if this content came from Anthropic
        if self.provider != "anthropic" or self.signature is None:
            return None

        result = {
            "type": "thinking",
            "thinking": self.text,
            "signature": self.signature,
        }

        if self.summary:
            result["summary"] = self.summary

        return result

    @classmethod
    def from_google(cls, part: Part) -> Optional["ThinkingContent"]:
        """Create ThinkingContent from Google Part object"""
        # vcprint(part, "Google Part", color="yellow")
        return cls(
            text=part.text or "",
            provider="google",
            signature=part.thought_signature if part.thought_signature else None,
        )

    @classmethod
    def from_anthropic(
        cls, content_block: dict[str, Any]
    ) -> Optional["ThinkingContent"]:
        """Create ThinkingContent from Anthropic content block"""
        return cls(
            text=content_block["thinking"],
            provider="anthropic",
            signature=content_block.get("signature"),
        )

    @classmethod
    def from_openai(
        cls, item: OpenAIResponseReasoningItem
    ) -> Optional["ThinkingContent"]:
        """Create ThinkingContent from OpenAI reasoning item"""
        # Extract encrypted_content from the item
        encrypted_content = getattr(item, "encrypted_content", None)

        return cls(
            summary=item.summary,
            id=item.id,
            provider="openai",
            signature=encrypted_content,
        )


# Union of all content types
UnifiedContent = (
    TextContent
    | ImageContent
    | AudioContent
    | VideoContent
    | YouTubeVideoContent
    | DocumentContent
    | ToolCallContent
    | ToolResultContent
    | ThinkingContent
    | CodeExecutionContent
    | CodeExecutionResultContent
    | WebSearchCallContent
)


def reconstruct_content(block: dict[str, Any]) -> UnifiedContent:
    """
    Reconstruct a content object from a stored JSONB block (cx_message.content).

    This is the deserialization counterpart to each content class's to_storage_dict() method.
    Handles all storage block types: text, thinking, media, tool_call, tool_result,
    code_exec, code_result, web_search.

    Args:
        block: A dict from the cx_message.content JSONB array with a 'type' discriminator.

    Returns:
        The appropriate UnifiedContent instance.
    """
    block_type = block.get("type", "text")

    if block_type == "text":
        metadata: dict[str, Any] = {}
        citations = block.get("citations")
        if citations:
            metadata["citations"] = citations
        return TextContent(
            text=block.get("text", ""),
            id=block.get("id", ""),
            metadata=metadata,
        )

    elif block_type == "thinking":
        import base64

        sig = block.get("signature")
        if sig and block.get("signature_encoding") == "base64":
            sig = base64.b64decode(sig)

        id_from_metadata = block.get("metadata", {}).get("id")
        return ThinkingContent(
            text=block.get("text", ""),
            provider=block.get("provider"),
            signature=sig,
            summary=block.get("summary", []),
            id=id_from_metadata or block.get("id", ""),
        )

    elif block_type == "media":
        result = reconstruct_media_content(block)
        if result is not None:
            return result
        return TextContent(text=f"[Unknown media kind: {block.get('kind')}]")

    elif block_type == "tool_call":
        return ToolCallContent(
            id=block.get("id", ""),
            call_id=block.get("call_id", ""),
            name=block.get("name", ""),
            arguments=block.get("arguments", {}),
        )

    elif block_type == "tool_result":
        return ToolResultContent(
            tool_use_id=block.get("tool_use_id", ""),
            call_id=block.get("call_id", ""),
            name=block.get("name", ""),
            content=block.get("content", ""),
            is_error=block.get("is_error", False),
        )

    elif block_type == "code_exec":
        return CodeExecutionContent(
            language=block.get("language", ""),
            code=block.get("code", ""),
        )

    elif block_type == "code_result":
        return CodeExecutionResultContent(
            output=block.get("output", ""),
            outcome=block.get("outcome", ""),
        )

    elif block_type == "web_search":
        meta = block.get("metadata", {})
        return WebSearchCallContent(
            id=block.get("id", ""),
            status=block.get("status", ""),
            action=meta.get("action", {}),
        )

    # Fallback: return as TextContent with warning
    vcprint(block, f"WARNING: Unknown storage block type: {block_type}", color="red")
    return TextContent(text=f"[Unknown block type: {block_type}]")



"""Browser session manager for persistent Playwright sessions across tool calls.

Sessions are keyed by session_id. Each session holds a live Playwright browser
and page. A model navigates with browser_navigate (which returns a session_id),
then uses that session_id for browser_click, browser_type_text, and
browser_screenshot to operate on the same live page.

Lifecycle:
- Sessions are created by browser_navigate and closed by browser_close.
- Idle sessions older than SESSION_TTL_SECONDS are evicted automatically by a
  background task that starts lazily when the first session is created.
- All sessions are force-closed on process shutdown via atexit.
"""
from __future__ import annotations

import asyncio
import atexit
import time
import uuid
from dataclasses import dataclass, field
from typing import Any

from matrx_utils import vcprint

SESSION_TTL_SECONDS = 300       # 5 minutes idle before auto-close
EVICTION_INTERVAL_SECONDS = 60  # check every minute


@dataclass
class BrowserSession:
    session_id: str
    pw: Any        # AsyncPlaywright instance
    browser: Any   # Browser instance
    page: Any      # Page instance
    last_used: float = field(default_factory=time.time)

    async def close(self) -> None:
        try:
            await self.browser.close()
        except Exception:
            pass
        try:
            await self.pw.stop()
        except Exception:
            pass


class BrowserSessionManager:
    """Process-global registry of live browser sessions with automatic TTL eviction."""

    def __init__(self) -> None:
        self._sessions: dict[str, BrowserSession] = {}
        self._lock = asyncio.Lock()
        self._eviction_task: asyncio.Task[None] | None = None

    def _ensure_eviction_task(self) -> None:
        """Start the background eviction task if not already running."""
        if self._eviction_task is None or self._eviction_task.done():
            try:
                loop = asyncio.get_running_loop()
                self._eviction_task = loop.create_task(self._eviction_loop(), name="browser-session-eviction")
            except RuntimeError:
                pass  # No running event loop yet — eviction starts on first create()

    async def _eviction_loop(self) -> None:
        while True:
            try:
                await asyncio.sleep(EVICTION_INTERVAL_SECONDS)
                await self.evict_stale()
            except asyncio.CancelledError:
                break
            except Exception as exc:
                vcprint(f"Browser session eviction error: {exc}", "[BrowserSessions] Eviction loop error", color="red")

    async def create(self) -> BrowserSession:
        from playwright.async_api import async_playwright

        pw = await async_playwright().start()
        browser = await pw.chromium.launch(headless=True)
        page = await browser.new_page()
        session_id = str(uuid.uuid4())[:8]
        session = BrowserSession(
            session_id=session_id,
            pw=pw,
            browser=browser,
            page=page,
        )
        async with self._lock:
            self._sessions[session_id] = session
        self._ensure_eviction_task()
        return session

    async def get(self, session_id: str) -> BrowserSession | None:
        async with self._lock:
            session = self._sessions.get(session_id)
            if session:
                session.last_used = time.time()
        return session

    async def close(self, session_id: str) -> None:
        async with self._lock:
            session = self._sessions.pop(session_id, None)
        if session:
            await session.close()

    async def close_all(self) -> None:
        async with self._lock:
            sessions = list(self._sessions.values())
            self._sessions.clear()
        if self._eviction_task and not self._eviction_task.done():
            self._eviction_task.cancel()
            try:
                await self._eviction_task
            except asyncio.CancelledError:
                pass
        for session in sessions:
            await session.close()

    async def evict_stale(self) -> None:
        cutoff = time.time() - SESSION_TTL_SECONDS
        async with self._lock:
            stale_ids = [sid for sid, s in self._sessions.items() if s.last_used < cutoff]
            stale_sessions = [self._sessions.pop(sid) for sid in stale_ids]
        for session in stale_sessions:
            await session.close()

    @property
    def active_count(self) -> int:
        return len(self._sessions)

    @property
    def session_ids(self) -> list[str]:
        return list(self._sessions.keys())


# ─── Process-level singleton ─────────────────────────────────────────────────

_manager: BrowserSessionManager | None = None


def get_browser_session_manager() -> BrowserSessionManager:
    global _manager
    if _manager is None:
        _manager = BrowserSessionManager()
    return _manager


def _shutdown_browser_sessions() -> None:
    """atexit handler — synchronously closes all sessions on process exit."""
    global _manager
    if _manager is None or _manager.active_count == 0:
        return
    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            loop.create_task(_manager.close_all())
        else:
            loop.run_until_complete(_manager.close_all())
    except Exception as exc:
        vcprint(f"Error during browser session shutdown: {exc}", "[BrowserSessions] Shutdown error", color="red")


atexit.register(_shutdown_browser_sessions)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class ToolCallUsage:
    """Tracks tool call information for a single request iteration."""

    iteration: int
    tool_calls_count: int
    """Number of tool calls made in this iteration"""

    tool_calls_details: list[dict[str, Any]]
    """List of tool call details (name, success, etc.)"""

    @staticmethod
    def aggregate(tool_call_list: list[ToolCallUsage]) -> dict[str, Any]:
        """Aggregate tool call statistics from a list of iteration tool calls.

        Returns a dictionary with comprehensive tool call analysis including:
        - Total tool calls across all iterations
        - Breakdown by tool name
        - Success/failure counts
        """
        if not tool_call_list:
            return {
                "total_tool_calls": 0,
                "iterations_with_tools": 0,
                "by_tool": {},
            }

        total_calls = sum(tc.tool_calls_count for tc in tool_call_list)
        iterations_with_tools = len(
            [tc for tc in tool_call_list if tc.tool_calls_count > 0]
        )

        # Aggregate by tool name
        by_tool: dict[str, dict[str, int]] = {}

        for tc in tool_call_list:
            for detail in tc.tool_calls_details:
                tool_name = detail.get("name", "unknown")
                if tool_name not in by_tool:
                    by_tool[tool_name] = {
                        "count": 0,
                        "success": 0,
                        "error": 0,
                    }

                by_tool[tool_name]["count"] += 1

                # Track success/error if available
                if detail.get("success", True):  # Default to True if not specified
                    by_tool[tool_name]["success"] += 1
                else:
                    by_tool[tool_name]["error"] += 1

        return {
            "total_tool_calls": total_calls,
            "iterations_with_tools": iterations_with_tools,
            "by_tool": by_tool,
        }


@dataclass
class TimingUsage:
    """Tracks timing information for a single request iteration."""

    start_time: float
    end_time: float | None = None

    # Component timings
    api_call_duration: float = 0.0
    tool_execution_duration: float = 0.0

    # Context
    model: str = ""
    iteration: int = 0

    @property
    def total_duration(self) -> float:
        """Total duration of this iteration in seconds."""
        if self.end_time and self.start_time:
            return self.end_time - self.start_time
        return 0.0

    @property
    def processing_duration(self) -> float:
        """Time spent processing logic outside of API and Tools."""
        return max(
            0.0,
            self.total_duration - self.api_call_duration - self.tool_execution_duration,
        )

    @staticmethod
    def aggregate(timings: list[TimingUsage]) -> dict[str, Any]:
        """Aggregate timing statistics from a list of iteration timings.

        Returns a dictionary with comprehensive timing analysis including:
        - Total wall-clock duration
        - Total time spent in API calls
        - Total time spent executing tools
        - Average times for each component
        """
        if not timings:
            return {
                "total_duration": 0.0,
                "api_duration": 0.0,
                "tool_duration": 0.0,
                "processing_duration": 0.0,
            }

        valid_timings = [t for t in timings if t.end_time is not None]
        if not valid_timings:
            return {
                "total_duration": 0.0,
                "api_duration": 0.0,
                "tool_duration": 0.0,
                "processing_duration": 0.0,
            }

        # Calculate totals
        total_api = sum(t.api_call_duration for t in valid_timings)
        total_tool = sum(t.tool_execution_duration for t in valid_timings)
        total_sum_duration = sum(t.total_duration for t in valid_timings)

        # Calculate wall clock duration (from start of first to end of last)
        # Sort just in case they aren't in order (though they should be)
        sorted_timings = sorted(valid_timings, key=lambda t: t.start_time)
        wall_clock_start = sorted_timings[0].start_time
        last_end_time = sorted_timings[-1].end_time
        assert last_end_time is not None  # guaranteed by valid_timings filter above
        wall_clock_duration = last_end_time - wall_clock_start

        return {
            "total_duration": wall_clock_duration,
            "sum_duration": total_sum_duration,
            "api_duration": total_api,
            "tool_duration": total_tool,
            "processing_duration": max(
                0.0, wall_clock_duration - total_api - total_tool
            ),
            "iterations": len(valid_timings),
            "avg_iteration_duration": wall_clock_duration / len(valid_timings)
            if valid_timings
            else 0,
        }



"""ai.tools — Unified tool execution system.

Public API:
    ToolExecutor              — The single entry point for all tool executions
    ToolRegistryV2            — Singleton registry (DB + code resolution)
    ToolDefinition            — Pydantic model describing a tool
    ToolContext               — Everything a tool needs about its execution environment
    ToolResult                — Structured result from any tool execution
    ToolError                 — Rich error with traceback + suggested action
    ToolType                  — LOCAL | EXTERNAL_MCP | AGENT | EXTERNAL_HANDLER
    CxToolCallRecord          — Pydantic model for the cx_tool_call DB row
    GuardrailEngine           — Centralized safety checks
    ToolStreamManager         — Streaming updates during tool execution
    ToolLifecycleManager      — Resource cleanup
    ToolExecutionLogger       — DB logging (two-phase: INSERT on start, UPDATE on finish)

External tool integration (for apps consuming the ai package):
    ExternalToolAdapter       — Base class: subclass, decorate with @external_tool, call register()
    external_tool             — Decorator: marks a method as the handler for a named tool
    ExternalHandlerRegistry   — Underlying singleton registry (advanced use / testing)
    ExternalToolHandler       — Type alias for the handler callable signature
    register_external_tool_handler  — Register a standalone function for one tool
    register_external_app_handler   — Register a catch-all for all tools from a source_app
"""

from .agent_tool import execute_agent_tool, register_agent_as_tool
from .executor import ToolExecutor
from .external_handlers import (
    ExternalHandlerRegistry,
    ExternalToolAdapter,
    ExternalToolHandler,
    external_tool,
    register_external_app_handler,
    register_external_tool_handler,
)
from .external_mcp import ExternalMCPClient
from .guardrails import GuardrailEngine
from .lifecycle import ToolLifecycleManager
from .logger import ToolExecutionLogger
from .models import (
    CxToolCallRecord,
    CxToolCallStatus,
    GuardrailResult,
    ToolContext,
    ToolDefinition,
    ToolError,
    ToolResult,
    ToolType,
)
from .registry import ToolRegistryV2
from .streaming import ToolStreamEvent, ToolStreamManager

__all__ = [
    "ToolExecutor",
    "ToolRegistryV2",
    "ToolDefinition",
    "ToolContext",
    "ToolResult",
    "ToolError",
    "ToolType",
    "CxToolCallRecord",
    "CxToolCallStatus",
    "GuardrailResult",
    "GuardrailEngine",
    "ToolStreamEvent",
    "ToolStreamManager",
    "ToolExecutionLogger",
    "ToolLifecycleManager",
    "ExternalMCPClient",
    "ExternalHandlerRegistry",
    "ExternalToolAdapter",
    "ExternalToolHandler",
    "external_tool",
    "register_external_tool_handler",
    "register_external_app_handler",
    "execute_agent_tool",
    "register_agent_as_tool",
]



"""Structured exception hierarchy and FastAPI exception handlers.

All exceptions that cross API boundaries should subclass MatrxException.
This guarantees a consistent error response shape across every service:

    {
        "error":  "<human-readable message>",
        "detail": { ... structured context ... },
        "path":   "/api/..."
    }

Every handler writes directly to sys.stderr before returning a response
so errors are never silenced by uvicorn log config or level filters.
"""

import sys
import traceback

from fastapi import HTTPException, Request
from fastapi.responses import ORJSONResponse

from matrx_service.context._log import log


class MatrxException(Exception):
    def __init__(
        self,
        message: str,
        status_code: int = 500,
        detail: dict | None = None,
    ) -> None:
        self.message = message
        self.status_code = status_code
        self.detail = detail or {}
        super().__init__(message)


class NotFoundError(MatrxException):
    def __init__(self, resource: str, resource_id: str) -> None:
        super().__init__(
            message=f"{resource} '{resource_id}' not found",
            status_code=404,
            detail={"resource": resource, "id": resource_id},
        )


class ProviderError(MatrxException):
    def __init__(self, provider: str, message: str, detail: dict | None = None) -> None:
        super().__init__(
            message=f"Provider '{provider}' error: {message}",
            status_code=502,
            detail={"provider": provider, **(detail or {})},
        )


class AgentNotFoundError(MatrxException):
    def __init__(self, agent_id: str) -> None:
        super().__init__(
            message=f"Agent '{agent_id}' not found",
            status_code=404,
            detail={"agent_id": agent_id},
        )


class ToolNotFoundError(MatrxException):
    def __init__(self, tool_name: str) -> None:
        super().__init__(
            message=f"Tool '{tool_name}' not found",
            status_code=404,
            detail={"tool_name": tool_name},
        )


class StreamingError(MatrxException):
    def __init__(self, message: str) -> None:
        super().__init__(message=f"Streaming error: {message}", status_code=500)


class AuthorizationError(MatrxException):
    def __init__(self, message: str = "Unauthorized") -> None:
        super().__init__(message=message, status_code=401, detail={"error": "unauthorized"})


class ForbiddenError(MatrxException):
    def __init__(self, message: str = "Forbidden") -> None:
        super().__init__(message=message, status_code=403, detail={"error": "forbidden"})


# ---------------------------------------------------------------------------
# Exception handlers — register all three in main.py
# ---------------------------------------------------------------------------


async def matrx_exception_handler(request: Request, exc: MatrxException) -> ORJSONResponse:
    print(
        f"\n[MatrxException] {request.method} {request.url.path}"
        f"\n  status={exc.status_code}  message={exc.message}"
        f"\n  detail={exc.detail}",
        file=sys.stderr,
        flush=True,
    )
    log(
        {"status_code": exc.status_code, "message": exc.message,
         "detail": exc.detail, "path": str(request.url.path)},
        title=f"[Exception] {request.method} {request.url.path}",
        color="red",
    )
    return ORJSONResponse(
        status_code=exc.status_code,
        content={
            "error": exc.message,
            "detail": exc.detail,
            "path": str(request.url.path),
        },
    )


async def http_exception_handler(request: Request, exc: HTTPException) -> ORJSONResponse:
    """Log every HTTPException to stderr so 401/403/404 are never silent."""
    detail = exc.detail
    print(
        f"\n[HTTPException] {request.method} {request.url.path}"
        f"\n  status={exc.status_code}  detail={detail}",
        file=sys.stderr,
        flush=True,
    )
    log(
        {"status_code": exc.status_code, "detail": detail, "path": str(request.url.path)},
        title=f"[HTTPException] {request.method} {request.url.path}",
        color="yellow",
    )
    return ORJSONResponse(
        status_code=exc.status_code,
        content={"error": detail, "path": str(request.url.path)},
        headers=dict(exc.headers) if exc.headers else {},
    )


async def unhandled_exception_handler(request: Request, exc: Exception) -> ORJSONResponse:
    tb = traceback.format_exc()
    print(
        f"\n[UNHANDLED EXCEPTION] {request.method} {request.url.path}"
        f"\n  {type(exc).__name__}: {exc}"
        f"\n{tb}",
        file=sys.stderr,
        flush=True,
    )
    log(
        tb,
        title=f"[Unhandled] {request.method} {request.url.path} | {type(exc).__name__}: {exc}",
        color="red",
    )
    return ORJSONResponse(
        status_code=500,
        content={
            "error": "Internal server error",
            "detail": {"type": type(exc).__name__, "message": str(exc)},
            "path": str(request.url.path),
        },
    )



# File: db/managers/workspaces.py
from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from matrx_orm import BaseManager, BaseDTO, ModelView, build_output_schema
from matrx_utils import vcprint

from db.models import Workspaces


# ---------------------------------------------------------------------------
# ModelView (new) — opt-in projection layer.
# Stores results flat on the model instance; no duplication, no nesting.
# To activate: set view_class = WorkspacesView on your manager subclass,
# or pass view_class=WorkspacesView to super().__init__().
# When active, the DTO path below is skipped automatically.
# ---------------------------------------------------------------------------

class WorkspacesView(ModelView[Workspaces]):
    """
    Declarative view for Workspaces.

    Configure what gets fetched and shaped automatically on every load:

      prefetch    — relation names to fetch concurrently (FK / IFK / M2M)
      exclude     — model field names to omit from to_dict() output
      inline_fk   — replace FK id columns with the full related object
                    e.g. {"customer_id": "customer"}

    Add async methods (no leading underscore) for computed fields:

        async def display_name(self, model: Workspaces) -> str:
            return model.name.title()
    """

    prefetch: list[str] = []
    exclude: list[str] = []
    inline_fk: dict[str, str] = {}

    # ------------------------------------------------------------------ #
    # Computed fields — add async methods below.                          #
    # Each method receives the model instance and returns a plain value.  #
    # Errors in computed fields are logged and stored as None —           #
    # they never abort the load.                                          #
    # ------------------------------------------------------------------ #


# ---------------------------------------------------------------------------
# Pydantic output schema (optional, requires pydantic v2).
# Auto-generated from the model's field definitions.  Useful for:
#   - FastAPI response_model type annotation
#   - JSON Schema generation: WorkspacesSchema.model_json_schema()
#   - Typed API responses: WorkspacesSchema.model_validate(item.to_dict())
#
# Usage example:
#   @app.get("/{id}", response_model=WorkspacesSchema)
#   async def get_workspaces(id: str):
#       item = await workspaces_manager_instance.load_by_id(id)
#       return item.to_dict()
# ---------------------------------------------------------------------------

try:
    WorkspacesSchema = build_output_schema(Workspaces)
except ImportError:
    WorkspacesSchema = None  # type: ignore[assignment]  # pydantic not installed


# ---------------------------------------------------------------------------
# BaseDTO (default) — active by default, fully backward compatible.
# Extend _process_core_data / _process_metadata with your business logic.
# When you are ready to migrate to the View above, set view_class on your
# manager subclass and this DTO will be bypassed automatically.
# ---------------------------------------------------------------------------

@dataclass
class WorkspacesDTO(BaseDTO[Workspaces]):
    id: str

    async def _initialize_dto(self, model: Workspaces) -> None:
        '''Override to populate DTO fields from the model.'''
        self.id = str(model.id)
        await self._process_core_data(model)
        await self._process_metadata(model)
        await self._initial_validation(model)
        self.initialized = True

    async def _process_core_data(self, model: Workspaces) -> None:
        '''Process core data from the model item.'''
        pass

    async def _process_metadata(self, model: Workspaces) -> None:
        '''Process metadata from the model item.'''
        pass

    async def _initial_validation(self, model: Workspaces) -> None:
        '''Validate fields from the model item.'''
        pass

    async def _final_validation(self) -> bool:
        '''Final validation of the model item.'''
        return True

    async def get_validated_dict(self) -> dict[str, Any]:
        '''Get the validated dictionary.'''
        await self._final_validation()
        return self.to_dict()


# ---------------------------------------------------------------------------
# Manager — DTO is active by default for full backward compatibility.
# To switch to the View (opt-in):
#   1. Quick: set view_class = WorkspacesView  (replaces DTO automatically)
#   2. Explicit: super().__init__(Workspaces, view_class=WorkspacesView)
# ---------------------------------------------------------------------------

class WorkspacesBase(BaseManager[Workspaces]):
    view_class = None  # DTO is used by default; set to WorkspacesView to opt in

    def __init__(
        self,
        dto_class: type[Any] | None = None,
        view_class: type[Any] | None = None,
    ) -> None:
        if view_class is not None:
            self.view_class = view_class
        super().__init__(Workspaces, dto_class=dto_class or WorkspacesDTO)

    def _initialize_manager(self) -> None:
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Workspaces) -> None:
        pass

    async def create_workspaces(self, **data: Any) -> Workspaces:
        return await self.create_item(**data)

    async def delete_workspaces(self, id: Any) -> bool:
        return await self.delete_item(id)

    async def get_workspaces_with_all_related(self, id: Any) -> tuple[Workspaces, Any]:
        return await self.get_item_with_all_related(id)

    async def load_workspaces_by_id(self, id: Any) -> Workspaces:
        return await self.load_by_id(id)

    async def load_workspaces(self, use_cache: bool = True, **kwargs: Any) -> Workspaces:
        return await self.load_item(use_cache, **kwargs)

    async def update_workspaces(self, id: Any, **updates: Any) -> Workspaces:
        return await self.update_item(id, **updates)

    async def load_workspaces(self, **kwargs: Any) -> list[Workspaces]:
        return await self.load_items(**kwargs)

    async def filter_workspaces(self, **kwargs: Any) -> list[Workspaces]:
        return await self.filter_items(**kwargs)

    async def get_or_create_workspaces(self, defaults: dict[str, Any] | None = None, **kwargs: Any) -> Workspaces | None:
        return await self.get_or_create(defaults, **kwargs)

    async def get_workspaces_with_organizations(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'organizations')

    async def get_workspaces_with_organizations(self) -> list[Any]:
        return await self.get_items_with_related('organizations')

    async def get_workspaces_with_self_reference(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'self_reference')

    async def get_workspaces_with_self_reference(self) -> list[Any]:
        return await self.get_items_with_related('self_reference')

    async def get_workspaces_with_context_variables(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'context_variables')

    async def get_workspaces_with_context_variables(self) -> list[Any]:
        return await self.get_items_with_related('context_variables')

    async def get_workspaces_with_user_active_context(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'user_active_context')

    async def get_workspaces_with_user_active_context(self) -> list[Any]:
        return await self.get_items_with_related('user_active_context')

    async def get_workspaces_with_broker_values(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'broker_values')

    async def get_workspaces_with_broker_values(self) -> list[Any]:
        return await self.get_items_with_related('broker_values')

    async def get_workspaces_with_user_files(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'user_files')

    async def get_workspaces_with_user_files(self) -> list[Any]:
        return await self.get_items_with_related('user_files')

    async def get_workspaces_with_projects(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'projects')

    async def get_workspaces_with_projects(self) -> list[Any]:
        return await self.get_items_with_related('projects')

    async def get_workspaces_with_cx_conversation(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'cx_conversation')

    async def get_workspaces_with_cx_conversation(self) -> list[Any]:
        return await self.get_items_with_related('cx_conversation')

    async def get_workspaces_with_sandbox_instances(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'sandbox_instances')

    async def get_workspaces_with_sandbox_instances(self) -> list[Any]:
        return await self.get_items_with_related('sandbox_instances')

    async def get_workspaces_with_context_items(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'context_items')

    async def get_workspaces_with_context_items(self) -> list[Any]:
        return await self.get_items_with_related('context_items')

    async def get_workspaces_with_app_instances(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'app_instances')

    async def get_workspaces_with_app_instances(self) -> list[Any]:
        return await self.get_items_with_related('app_instances')

    async def get_workspaces_with_notes(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'notes')

    async def get_workspaces_with_notes(self) -> list[Any]:
        return await self.get_items_with_related('notes')

    async def get_workspaces_with_workspace_members(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'workspace_members')

    async def get_workspaces_with_workspace_members(self) -> list[Any]:
        return await self.get_items_with_related('workspace_members')

    async def get_workspaces_with_workspace_invitations(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'workspace_invitations')

    async def get_workspaces_with_workspace_invitations(self) -> list[Any]:
        return await self.get_items_with_related('workspace_invitations')

    async def get_workspaces_with_user_tables(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'user_tables')

    async def get_workspaces_with_user_tables(self) -> list[Any]:
        return await self.get_items_with_related('user_tables')

    async def get_workspaces_with_transcripts(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'transcripts')

    async def get_workspaces_with_transcripts(self) -> list[Any]:
        return await self.get_items_with_related('transcripts')

    async def get_workspaces_with_workflow(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'workflow')

    async def get_workspaces_with_workflow(self) -> list[Any]:
        return await self.get_items_with_related('workflow')

    async def get_workspaces_with_canvas_items(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'canvas_items')

    async def get_workspaces_with_canvas_items(self) -> list[Any]:
        return await self.get_items_with_related('canvas_items')

    async def get_workspaces_with_prompts(self, id: Any) -> tuple[Any, Any]:
        return await self.get_item_with_related(id, 'prompts')

    async def get_workspaces_with_prompts(self) -> list[Any]:
        return await self.get_items_with_related('prompts')

    async def load_workspaces_by_organization_id(self, organization_id: Any) -> list[Any]:
        return await self.load_items(organization_id=organization_id)

    async def filter_workspaces_by_organization_id(self, organization_id: Any) -> list[Any]:
        return await self.filter_items(organization_id=organization_id)

    async def load_workspaces_by_parent_workspace_id(self, parent_workspace_id: Any) -> list[Any]:
        return await self.load_items(parent_workspace_id=parent_workspace_id)

    async def filter_workspaces_by_parent_workspace_id(self, parent_workspace_id: Any) -> list[Any]:
        return await self.filter_items(parent_workspace_id=parent_workspace_id)

    async def load_workspaces_by_created_by(self, created_by: Any) -> list[Any]:
        return await self.load_items(created_by=created_by)

    async def filter_workspaces_by_created_by(self, created_by: Any) -> list[Any]:
        return await self.filter_items(created_by=created_by)

    async def load_workspaces_by_ids(self, ids: list[Any]) -> list[Any]:
        return await self.load_items_by_ids(ids)

    def add_computed_field(self, field: str) -> None:
        super().add_computed_field(field)

    def add_relation_field(self, field: str) -> None:
        super().add_relation_field(field)

    @property
    def active_workspaces_ids(self) -> set[Any]:
        return self.active_item_ids



class WorkspacesManager(WorkspacesBase):
    _instance: WorkspacesManager | None = None

    def __new__(cls, *args: Any, **kwargs: Any) -> WorkspacesManager:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        super().__init__()

    async def _initialize_runtime_data(self, item: Workspaces) -> None:
        pass

workspaces_manager_instance = WorkspacesManager()


"""
Transcription Cache

Simple in-memory cache for audio transcriptions to avoid re-transcribing
the same audio files across different requests.

Cache is keyed by (audio_source, model, language) to ensure we only reuse
transcriptions that used the same settings.
"""
from __future__ import annotations

import hashlib
from dataclasses import dataclass
from typing import Any


@dataclass
class CachedTranscription:
    """Cached transcription result"""
    
    text: str
    """The transcribed text"""
    
    metadata: dict[str, Any]
    """Metadata from the transcription (usage, quality metrics, etc.)"""

class TranscriptionCache:
    """
    Simple in-memory cache for transcriptions.
    
    Caches transcriptions by (audio_source, model, language) to avoid
    re-transcribing the same audio with the same settings.
    """
    
    def __init__(self):
        self._cache: dict[str, CachedTranscription] = {}
    
    def _generate_key(
        self, 
        audio_source: str, 
        model: str, 
        language: str | None
    ) -> str:
        """
        Generate cache key from audio source, model, and language.
        
        Uses hash of the audio source for reasonable key length.
        """
        # Hash the audio source to create a reasonable key
        source_hash = hashlib.md5(audio_source.encode()).hexdigest()[:16]
        
        # Combine with model and language
        lang_str = language or "auto"
        return f"{source_hash}:{model}:{lang_str}"
    
    def get(
        self, 
        audio_source: str, 
        model: str, 
        language: str | None
    ) -> CachedTranscription | None:
        """
        Get cached transcription if available.
        
        Args:
            audio_source: URL, file path, or base64 data
            model: Whisper model name
            language: Language code (or None for auto-detect)
        
        Returns:
            CachedTranscription if found, None otherwise
        """
        key = self._generate_key(audio_source, model, language)
        return self._cache.get(key)
    
    def set(
        self,
        audio_source: str,
        model: str,
        language: str | None,
        text: str,
        metadata: dict[str, Any]
    ) -> None:
        """
        Cache a transcription result.
        
        Args:
            audio_source: URL, file path, or base64 data
            model: Whisper model name
            language: Language code (or None for auto-detect)
            text: Transcribed text
            metadata: Transcription metadata (usage, quality metrics, etc.)
        """
        key = self._generate_key(audio_source, model, language)
        self._cache[key] = CachedTranscription(text=text, metadata=metadata)
    
    def clear(self) -> None:
        """Clear all cached transcriptions"""
        self._cache.clear()
    
    def size(self) -> int:
        """Get number of cached transcriptions"""
        return len(self._cache)

# Global cache instance
_global_cache = TranscriptionCache()

def get_cache() -> TranscriptionCache:
    """Get the global transcription cache instance"""
    return _global_cache

def clear_cache() -> None:
    """Clear the global transcription cache"""
    _global_cache.clear()



from __future__ import annotations

import traceback
from datetime import date, timedelta
from typing import Any

from matrx_ai.tools.models import ToolContext, ToolError, ToolResult
from matrx_ai.tools.output_models.seo import (
    SeoKeywordDataOutput,
    normalize_keyword_item,
)


async def seo_check_meta_titles(args: dict[str, Any], ctx: ToolContext) -> ToolResult:
    from seo.utils.meta_calculators import calculate_meta_title_metrics

    titles = args.get("titles", [])
    if isinstance(titles, str):
        titles = [titles]
    if not titles:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="titles must be a non-empty string or list of strings.",
            ),
        )

    try:
        analysis = []
        for title in titles:
            metrics = calculate_meta_title_metrics(title)
            analysis.append(
                {
                    "title": title,
                    "pixel_width": metrics.get("pixel_width"),
                    "character_count": metrics.get("character_count"),
                    "desktop_ok": metrics.get("desktop_ok"),
                    "mobile_ok": metrics.get("mobile_ok"),
                    "seo_length_ok": metrics.get("seo_length_ok"),
                    "too_short": metrics.get("too_short"),
                    "issues": metrics.get("issues", []),
                    "title_ok": metrics.get("title_ok"),
                }
            )
        return ToolResult(
            success=True, output={"title_analysis": analysis, "count": len(analysis)}
        )
    except Exception as e:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="execution", message=str(e), traceback=traceback.format_exc()
            ),
        )


async def seo_check_meta_descriptions(
    args: dict[str, Any], ctx: ToolContext
) -> ToolResult:
    from seo.utils.meta_calculators import calculate_meta_description_metrics

    descriptions = args.get("descriptions", [])
    if isinstance(descriptions, str):
        descriptions = [descriptions]
    if not descriptions:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="descriptions must be a non-empty string or list of strings.",
            ),
        )

    try:
        analysis = []
        for desc in descriptions:
            metrics = calculate_meta_description_metrics(desc)
            analysis.append(
                {
                    "description": desc,
                    "pixel_width": metrics.get("pixel_width"),
                    "character_count": metrics.get("character_count"),
                    "desktop_ok": metrics.get("desktop_ok"),
                    "mobile_ok": metrics.get("mobile_ok"),
                    "seo_length_ok": metrics.get("seo_length_ok"),
                    "too_short": metrics.get("too_short"),
                    "issues": metrics.get("issues", []),
                    "description_ok": metrics.get("description_ok"),
                }
            )
        return ToolResult(
            success=True,
            output={"description_analysis": analysis, "count": len(analysis)},
        )
    except Exception as e:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="execution", message=str(e), traceback=traceback.format_exc()
            ),
        )


async def seo_check_meta_tags_batch(
    args: dict[str, Any], ctx: ToolContext
) -> ToolResult:
    from seo.utils.meta_calculators import analyze_meta_tags_batch

    meta_data = args.get("meta_data", [])
    if not meta_data or not isinstance(meta_data, list):
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="meta_data must be a non-empty list of objects with 'title' and/or 'description'.",
            ),
        )

    for idx, item in enumerate(meta_data):
        if not isinstance(item, dict) or (
            "title" not in item and "description" not in item
        ):
            return ToolResult(
                success=False,
                error=ToolError(
                    error_type="validation",
                    message=f"Item at index {idx} must have at least 'title' or 'description'.",
                ),
            )

    try:
        result = analyze_meta_tags_batch(meta_data)
        return ToolResult(
            success=True, output={"batch_analysis": result, "count": len(meta_data)}
        )
    except Exception as e:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="execution", message=str(e), traceback=traceback.format_exc()
            ),
        )


async def seo_get_keyword_data(args: dict[str, Any], ctx: ToolContext) -> ToolResult:
    import json

    from api_management.api_sdks.data_for_seo.data_for_seo import keyword_research

    keywords = args.get("keywords", [])
    date_from_str = args.get("date_from", "")
    date_to_str = args.get("date_to", "")
    location_code = args.get("location_code", 2840)
    language_code = args.get("language_code", "en")
    search_partners = args.get("search_partners", True)
    sort_by = args.get("sort_by", "search_volume")

    if not keywords or not isinstance(keywords, list):
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="keywords must be a non-empty list of strings.",
            ),
        )
    if not date_from_str or not date_to_str:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="date_from and date_to are required (YYYY-MM-DD).",
            ),
        )

    try:
        d_from = date.fromisoformat(date_from_str.strip())
        d_to = date.fromisoformat(date_to_str.strip())
    except ValueError:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="date_from and date_to must be valid dates in YYYY-MM-DD format.",
            ),
        )

    today = date.today()
    if d_from > today or d_to > today:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="date_from and date_to must be in the past. Google Ads does not return data for future dates.",
                suggested_action="Use past dates only, e.g. date_from='2024-01-01', date_to='2024-12-31'.",
            ),
        )
    if d_from > d_to:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message="date_from must be before or equal to date_to.",
                suggested_action="Ensure date_from is the start of the range and date_to is the end, e.g. date_from='2024-01-01', date_to='2024-12-31'.",
            ),
        )
    four_years_ago = today - timedelta(days=4 * 365)
    if d_from < four_years_ago:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="validation",
                message=f"date_from cannot be older than 4 years. Minimum allowed: {four_years_ago.isoformat()}.",
                suggested_action=f"Use date_from >= {four_years_ago.isoformat()}.",
            ),
        )

    date_from = date_from_str.strip()
    date_to = date_to_str.strip()

    try:
        response = keyword_research(
            keywords=keywords,
            date_from=date_from,
            date_to=date_to,
            location_code=location_code,
            language_code=language_code,
            search_partners=search_partners,
            sort_by=sort_by,
        )
        data = json.loads(response) if isinstance(response, str) else response

        tasks = data.get("tasks", [])
        if not tasks or tasks[0].get("status_code") != 20000:
            return ToolResult(
                success=False,
                error=ToolError(
                    error_type="api_error",
                    message=f"DataForSEO API error: {tasks[0].get('status_message', 'Unknown') if tasks else 'No tasks returned'}",
                ),
            )

        result_data = tasks[0].get("result", []) or []
        keywords_data = [
            normalize_keyword_item(item)
            for item in result_data
            if isinstance(item, dict)
        ]
        output = SeoKeywordDataOutput(
            keywords_data=keywords_data,
            total_keywords=len(keywords),
            date_range={"from": date_from, "to": date_to},
            search_parameters={
                "location_code": location_code,
                "language_code": language_code,
            },
        )
        return ToolResult(success=True, output=output.model_dump())
    except Exception as e:
        return ToolResult(
            success=False,
            error=ToolError(
                error_type="execution", message=str(e), traceback=traceback.format_exc()
            ),
        )



# File: db/helpers/auto_config.py
ai_provider_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'AiProvider', 'model_name': 'ai_provider', 'model_name_plural': 'ai_providers', 'model_name_snake': 'ai_provider', 'relations': ['ai_settings', 'ai_model'], 'filter_fields': [], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_agent_memory_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxAgentMemory', 'model_name': 'cx_agent_memory', 'model_name_plural': 'cx_agent_memories', 'model_name_snake': 'cx_agent_memory', 'relations': [], 'filter_fields': ['user_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


organizations_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'Organizations', 'model_name': 'organizations', 'model_name_plural': 'organizations', 'model_name_snake': 'organizations', 'relations': ['workspaces', 'context_variables', 'permissions', 'organization_members', 'user_active_context', 'broker_values', 'user_files', 'projects', 'organization_invitations', 'cx_conversation', 'sandbox_instances', 'context_items', 'app_instances', 'notes', 'user_tables', 'transcripts', 'workflow', 'canvas_items', 'prompts'], 'filter_fields': ['created_by'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


shortcut_categories_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'ShortcutCategories', 'model_name': 'shortcut_categories', 'model_name_plural': 'shortcut_categories', 'model_name_snake': 'shortcut_categories', 'relations': ['self_reference', 'content_blocks', 'prompt_shortcuts', 'system_prompts_new'], 'filter_fields': ['parent_category_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


tools_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'Tools', 'model_name': 'tools', 'model_name_plural': 'tools', 'model_name_snake': 'tools', 'relations': ['tool_versions', 'tool_test_samples', 'tool_ui_components'], 'filter_fields': ['tags'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


ai_model_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'AiModel', 'model_name': 'ai_model', 'model_name_plural': 'ai_models', 'model_name_snake': 'ai_model', 'relations': ['ai_provider', 'ai_model_endpoint', 'ai_settings', 'recipe_model'], 'filter_fields': ['name', 'common_name', 'provider', 'model_class', 'model_provider'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


content_blocks_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'ContentBlocks', 'model_name': 'content_blocks', 'model_name_plural': 'content_blocks', 'model_name_snake': 'content_blocks', 'relations': ['shortcut_categories'], 'filter_fields': ['category_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


workspaces_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'Workspaces', 'model_name': 'workspaces', 'model_name_plural': 'workspaces', 'model_name_snake': 'workspaces', 'relations': ['organizations', 'self_reference', 'context_variables', 'user_active_context', 'broker_values', 'user_files', 'projects', 'cx_conversation', 'sandbox_instances', 'context_items', 'app_instances', 'notes', 'workspace_members', 'workspace_invitations', 'user_tables', 'transcripts', 'workflow', 'canvas_items', 'prompts'], 'filter_fields': ['organization_id', 'parent_workspace_id', 'created_by'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


projects_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'Projects', 'model_name': 'projects', 'model_name_plural': 'projects', 'model_name_snake': 'projects', 'relations': ['organizations', 'workspaces', 'context_variables', 'tasks', 'user_active_context', 'broker_values', 'user_files', 'cx_conversation', 'project_members', 'project_invitations', 'sandbox_instances', 'context_items', 'app_instances', 'notes', 'rs_topic', 'user_tables', 'transcripts', 'workflow', 'canvas_items', 'prompts'], 'filter_fields': ['created_by', 'organization_id', 'workspace_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


tasks_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'Tasks', 'model_name': 'tasks', 'model_name_plural': 'tasks', 'model_name_snake': 'tasks', 'relations': ['self_reference', 'projects', 'task_assignments', 'task_attachments', 'context_variables', 'task_comments', 'user_active_context', 'broker_values', 'user_files', 'cx_conversation', 'sandbox_instances', 'context_items', 'app_instances', 'notes', 'user_tables', 'transcripts', 'workflow', 'prompts'], 'filter_fields': ['project_id', 'user_id', 'parent_task_id', 'priority', 'assignee_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_conversation_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxConversation', 'model_name': 'cx_conversation', 'model_name_plural': 'cx_conversations', 'model_name_snake': 'cx_conversation', 'relations': ['self_reference', 'ai_model', 'organizations', 'projects', 'tasks', 'workspaces', 'cx_tool_call', 'cx_message', 'cx_media', 'cx_user_request', 'cx_request', 'canvas_items'], 'filter_fields': ['user_id', 'forked_from_id', 'last_model_id', 'parent_conversation_id', 'keywords', 'organization_id', 'workspace_id', 'project_id', 'task_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


prompts_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'Prompts', 'model_name': 'prompts', 'model_name_plural': 'prompts', 'model_name_snake': 'prompts', 'relations': ['ai_model', 'organizations', 'projects', 'tasks', 'workspaces', 'system_prompts_new', 'prompt_actions', 'system_prompts', 'prompt_versions', 'prompt_builtins', 'prompt_apps'], 'filter_fields': ['user_id', 'tags', 'model_id', 'organization_id', 'workspace_id', 'project_id', 'task_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


user_tables_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'UserTables', 'model_name': 'user_tables', 'model_name_plural': 'user_tables', 'model_name_snake': 'user_tables', 'relations': ['organizations', 'projects', 'tasks', 'workspaces', 'table_data', 'table_fields'], 'filter_fields': ['user_id', 'organization_id', 'workspace_id', 'project_id', 'task_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_media_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxMedia', 'model_name': 'cx_media', 'model_name_plural': 'cx_media', 'model_name_snake': 'cx_media', 'relations': ['cx_conversation'], 'filter_fields': ['conversation_id', 'user_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_message_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxMessage', 'model_name': 'cx_message', 'model_name_plural': 'cx_messages', 'model_name_snake': 'cx_message', 'relations': ['cx_conversation', 'cx_tool_call'], 'filter_fields': ['conversation_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_user_request_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxUserRequest', 'model_name': 'cx_user_request', 'model_name_plural': 'cx_user_requests', 'model_name_snake': 'cx_user_request', 'relations': ['cx_conversation', 'cx_tool_call', 'cx_request'], 'filter_fields': ['conversation_id', 'user_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


prompt_builtins_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'PromptBuiltins', 'model_name': 'prompt_builtins', 'model_name_plural': 'prompt_builtins', 'model_name_snake': 'prompt_builtins', 'relations': ['ai_model', 'prompts', 'prompt_shortcuts', 'prompt_actions', 'prompt_builtin_versions'], 'filter_fields': ['created_by_user_id', 'source_prompt_id', 'tags', 'model_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


table_data_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'TableData', 'model_name': 'table_data', 'model_name_plural': 'table_data', 'model_name_snake': 'table_data', 'relations': ['user_tables'], 'filter_fields': ['table_id', 'user_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_request_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxRequest', 'model_name': 'cx_request', 'model_name_plural': 'cx_requests', 'model_name_snake': 'cx_request', 'relations': ['ai_model', 'cx_conversation', 'cx_user_request'], 'filter_fields': ['user_request_id', 'conversation_id', 'ai_model_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


cx_tool_call_auto_config = {'models_module_path': 'db.models', 'model_pascal': 'CxToolCall', 'model_name': 'cx_tool_call', 'model_name_plural': 'cx_tool_calls', 'model_name_snake': 'cx_tool_call', 'relations': ['cx_conversation', 'cx_message', 'cx_user_request', 'self_reference'], 'filter_fields': ['conversation_id', 'message_id', 'user_id', 'user_request_id', 'parent_call_id'], 'include_core_relations': True, 'include_active_relations': False, 'include_filter_fields': True, 'include_active_methods': False, 'include_or_not_methods': False, 'include_to_dict_methods': False, 'include_to_dict_relations': False, 'm2m_relations': None, 'view_prefetch': []}


from dotenv import load_dotenv
from matrx_utils import clear_terminal, vcprint
from openai import OpenAI
from openai.types.responses import Response as OpenAIResponse
from openai.types.responses import ResponseOutputItem as OpenAIResponseOutputItem

from matrx_ai.config.unified_config import UnifiedMessage

load_dotenv()
clear_terminal()

client = OpenAI()

# Use the streaming context manager to get both events and final response
with client.responses.stream(
    model="gpt-5-mini",
    reasoning={
        "effort": "medium",
        "summary": "detailed"
    },
    tools=[{ "type": "web_search_preview" }],
    include=["reasoning.encrypted_content"],
    input="What was a positive news story from today?",
) as stream:
    # Process streaming events in real-time
    for event in stream:
        event_type = event.type
        
        # Handle different event types for streaming display
        if event_type == "response.output_text.delta":
            print(event.delta, end="", flush=True)
        elif event_type == "response.reasoning_summary_text.delta":
            print(f"{event.delta}", end="", flush=True)
        elif event_type == "response.created":
            vcprint("Stream started", color="cyan")
        elif event_type == "response.completed":
            vcprint("\nStream completed", color="green")
    
    # Get the final complete response with full content and usage data
    response: OpenAIResponse = stream.get_final_response()

print("\n" + "="*50)
vcprint(response, "Final Complete Response", color="green")


# Output is a list of items that can be messages, reasoning, web_search_call, etc.
output: list[OpenAIResponseOutputItem] = response.output

for item in output:
    item_type: str = item.type
    if item_type == "message":
        pass
    elif item_type == "reasoning":
        pass
    elif item_type == "web_search_call":
        pass
    else:
        vcprint(item, "Unknown Item", color="red")

unified_message: UnifiedMessage = UnifiedMessage.from_openai(output)
vcprint(unified_message, "Unified Message", color="magenta")

sample_output: list[OpenAIResponseOutputItem] = [
        {
            "id": "ws_0a12a58b7b1fca6000695550b609b88193a3323f0785901b3d",
            "action": {
                "query": "positive news story today",
                "type": "search",
                "sources": None
            },
            "status": "completed",
            "type": "web_search_call"
        },
        {
            "id": "msg_0a12a58b7b1fca6000695550b710b081938dae0143aaf39048",
            "content": [
                {
                    "annotations": [
                        {
                            "end_index": 495,
                            "start_index": 303,
                            "title": "School assembly news headlines, December 31: Top India, world, sports, business news - India Today",
                            "type": "url_citation",
                            "url": "https://www.indiatoday.in/amp/education-today/news/story/school-assembly-news-headlines-december-31-top-india-world-sports-business-news-2844210-2025-12-31?utm_source=openai"
                        },
                        {
                            "end_index": 950,
                            "start_index": 865,
                            "title": "2025 in science",
                            "type": "url_citation",
                            "url": "https://en.wikipedia.org/wiki/2025_in_science?utm_source=openai"
                        }
                    ],
                    "text": "On December 31, 2025, a notable positive development was the strong performance of global stock markets. European indices reached record highs, and there were significant gains in both Asian and U.S. markets. This resilience reflects investor confidence and sets a promising tone for the upcoming year. ([indiatoday.in](https://www.indiatoday.in/amp/education-today/news/story/school-assembly-news-headlines-december-31-top-india-world-sports-business-news-2844210-2025-12-31?utm_source=openai))\n\nAdditionally, in the realm of science, UK company Space Forge achieved a milestone in space-based manufacturing by successfully demonstrating a 1,000 \u00b0C furnace aboard an orbital micro-factory. This advancement is designed to produce semiconductor materials in microgravity, potentially resulting in materials up to 4,000 times purer than those manufactured on Earth. ([en.wikipedia.org](https://en.wikipedia.org/wiki/2025_in_science?utm_source=openai))\n\nThese developments highlight significant progress in both economic and technological fields as the year concludes. ",
                    "type": "output_text",
                    "logprobs": []
                }
            ],
            "role": "assistant",
            "status": "completed",
            "type": "message"
        }
    ]

full_response_example: OpenAIResponse = {
    "id": "resp_0a16d8c172123aeb006955587f41f481a0b884a4f06bedc90c",
    "created_at": 1767200895.0,
    "error": None,
    "incomplete_details": None,
    "instructions": None,
    "metadata": {},
    "model": "gpt-5.1-2025-11-13",
    "object": "response",
    "output": [
        {
            "id": "rs_0a16d8c172123aeb006955587ff73081a099639e7f57bdd432",
            "summary": [
                {
                    "text": "**Searching for positive news**\n\nI need to find a positive news story from today, December 31, 2025. The user is looking for something uplifting, possibly related to science, the environment, or medicine. I'll use the web tool to execute a search query focusing on today's date for recent and relevant news. I\u2019ll make sure to pull just one great story that fits what the user is interested in! Let's see what we can find.",
                    "type": "summary_text"
                }
            ],
            "type": "reasoning",
            "content": None,
            "encrypted_content": None,
            "status": None
        },
        {
            "id": "ws_0a16d8c172123aeb006955588402c481a096d2645bee7a9be5",
            "action": {
                "query": "good news positive news December 31 2025 breakthrough discovery",
                "type": "search",
                "sources": None
            },
            "status": "completed",
            "type": "web_search_call"
        },
        {
            "id": "rs_0a16d8c172123aeb00695558935a0881a0bf1a647cc846ee99",
            "summary": [
                {
                    "text": "**Searching for positive news**\n\nI need to find a positive story specifically from today, December 31, 2025. I've noticed some general \"year in review\" articles and positive science updates, but the user asked for something published today. I should refine my search for \"December 31, 2025 good news.\" I\u2019ll aim for a mainstream piece, possibly something like an article about a malaria vaccine cost drop or conservation success. I\u2019ll focus on getting a concrete story.",
                    "type": "summary_text"
                }
            ],
            "type": "reasoning",
            "content": None,
            "encrypted_content": None,
            "status": None
        },
        {
            "id": "ws_0a16d8c172123aeb00695558953d2081a0b5781268222e73c8",
            "action": {
                "query": "\"December 31, 2025\" \"good news\" \"breakthrough\"",
                "type": "search",
                "sources": None
            },
            "status": "completed",
            "type": "web_search_call"
        },
        {
            "id": "rs_0a16d8c172123aeb00695558a09c5881a0a02af404b0d209f2",
            "summary": [
                {
                    "text": "**Summarizing conservation success**\n\nThere\u2019s a recent article titled \"DEC completes spruce grouse restoration effort\" published on December 31, 2025, detailing the successful conservation of the endangered spruce grouse in New York. This counts as positive news since it highlights a successful effort in wildlife restoration in the user's state. I should summarize the key points briefly, providing just enough context to explain why it\u2019s a good development, and cite the source as FingerLakes1.com.",
                    "type": "summary_text"
                },
                {
                    "text": "**Finalizing spruce grouse summary**\n\nI\u2019m reviewing the details of the article to ensure I have everything right. The New York State Department of Environmental Conservation has completed its 13-year spruce grouse recovery program in the Adirondacks, releasing a total of 363 adults and 711 chicks. Their final release included 50 adults in summer 2025, and now they're focusing on monitoring and habitat management. This is a major milestone for conservation, showing a shift from decline to hope for this endangered species.",
                    "type": "summary_text"
                },
                {
                    "text": "**Defining spruce grouse**\n\nI'll quickly define the spruce grouse as a rare bird found in cold forests, situated near the southern edge of its range in New York. This is positive news as it represents a long-term investment in wildlife conservation and showcases cross-border cooperation efforts for species protection. Overall, this achievement is significant for restoring populations. I'll make sure to include the explicit date of December 31, 2025, in my response.",
                    "type": "summary_text"
                }
            ],
            "type": "reasoning",
            "content": None,
            "encrypted_content": None,
            "status": None
        },
        {
            "id": "msg_0a16d8c172123aeb00695558a8999481a0ba1cf75662ed604e",
            "content": [
                {
                    "annotations": [
                        {
                            "end_index": 796,
                            "start_index": 669,
                            "title": "DEC completes spruce grouse restoration effort",
                            "type": "url_citation",
                            "url": "https://www.fingerlakes1.com/2025/12/31/dec-completes-spruce-grouse-restoration-effort/?utm_source=openai"
                        }
                    ],
                    "text": "One uplifting story from today (December 31, 2025):\n\nNew York\u2019s Department of Environmental Conservation announced it has completed the release phase of a 13\u2011year program to rescue the endangered spruce grouse, one of the state\u2019s rarest birds, in the Adirondacks. \n\nSince 2013, biologists and partners in Canada and Maine have moved hundreds of adult birds and chicks into carefully selected forest habitats to rebuild the population. With the final group of birds released this past summer, officials say the focus can now shift from emergency releases to long\u2011term monitoring and habitat protection\u2014meaning the species is no longer in immediate freefall in New York. ([fingerlakes1.com](https://www.fingerlakes1.com/2025/12/31/dec-completes-spruce-grouse-restoration-effort/?utm_source=openai))\n\nSo, today\u2019s good news: a long, patient conservation effort has reached a major milestone, giving a threatened species a real chance to recover.",
                    "type": "output_text",
                    "logprobs": []
                }
            ],
            "role": "assistant",
            "status": "completed",
            "type": "message"
        }
    ],
    "parallel_tool_calls": True,
    "temperature": 1.0,
    "tool_choice": "auto",
    "tools": [
        {
            "type": "web_search_preview",
            "search_context_size": "medium",
            "user_location": {
                "type": "approximate",
                "city": None,
                "country": "US",
                "region": None,
                "timezone": None
            }
        }
    ],
    "top_p": 1.0,
    "background": False,
    "conversation": None,
    "max_output_tokens": None,
    "max_tool_calls": None,
    "previous_response_id": None,
    "prompt": None,
    "prompt_cache_key": None,
    "reasoning": {
        "effort": "high",
        "generate_summary": None,
        "summary": "detailed"
    },
    "safety_identifier": None,
    "service_tier": "default",
    "status": "completed",
    "text": {
        "format": {
            "type": "text"
        },
        "verbosity": "medium"
    },
    "top_logprobs": 0,
    "truncation": "disabled",
    "usage": {
        "input_tokens": 12296,
        "input_tokens_details": {
            "cached_tokens": 0
        },
        "output_tokens": 1076,
        "output_tokens_details": {
            "reasoning_tokens": 893
        },
        "total_tokens": 13372
    },
    "user": None,
    "billing": {
        "payer": "developer"
    },
    "completed_at": 1767200936,
    "prompt_cache_retention": None,
    "store": True
}


from typing import Any

from pydantic import BaseModel, Field


class ToolParameter(BaseModel):
    name: str
    type: str
    description: str
    required: bool = True
    default: Any = None


class ToolDefinition(BaseModel):
    name: str
    description: str
    parameters: list[ToolParameter] = Field(default_factory=list)
    metadata: dict[str, Any] = Field(default_factory=dict)


class ToolCallRequest(BaseModel):
    tool_name: str
    arguments: dict[str, Any] = Field(default_factory=dict)
    call_id: str | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)


class ToolCallResult(BaseModel):
    call_id: str | None = None
    tool_name: str
    result: Any
    error: str | None = None
    elapsed_ms: float | None = None



from matrx_ai.agent_runners.research import (
    AgentResult,
    scrape_research_condenser_agent_1,
    scrape_research_condenser_agent_2,
)

__all__ = [
    "AgentResult",
    "scrape_research_condenser_agent_1",
    "scrape_research_condenser_agent_2",
]



from __future__ import annotations

import asyncio
import os
import traceback
from collections.abc import AsyncIterator

import rich
from google import genai
from google.genai.types import (
    Candidate,
    GenerateContentResponse,
    Part,
)
from matrx_utils import vcprint

from matrx_ai.config import (
    UnifiedConfig,
    UnifiedResponse,
)
from matrx_ai.context.emitter_protocol import Emitter

from .translator import GoogleProviderConfig, GoogleTranslator

LOCAL_DEBUG = False

class GoogleChat:
    """Google Gemini-specific endpoint implementation."""

    client: genai.Client
    endpoint_name: str
    translator: GoogleTranslator
    debug: bool

    def __init__(self):
        self.client = genai.Client(api_key=os.environ.get("GEMINI_API_KEY"))
        self.endpoint_name = "[GOOGLE CHAT]"
        self.translator = GoogleTranslator()
        self.debug = LOCAL_DEBUG

    async def execute(
        self,
        unified_config: UnifiedConfig,
        api_class: str,
        debug: bool = False,
    ) -> UnifiedResponse:
        from matrx_ai.context.app_context import get_app_context

        emitter = get_app_context().emitter

        config_data: GoogleProviderConfig = self.translator.to_google(
            unified_config, api_class
        )
        self.debug = debug or LOCAL_DEBUG
        matrx_model_name = unified_config.model

        vcprint(f"[Google Chat] executing, with debug: {self.debug}", color="blue")
        # rich.print(config_data)

        if self.debug:
            rich.print(config_data)

        try:
            if unified_config.stream:
                response: AsyncIterator[GenerateContentResponse] = (
                    self.client.models.generate_content_stream(**config_data)
                )

                accumulated_chunks: list[GenerateContentResponse] = []

                chunk: GenerateContentResponse
                for chunk in response:
                    accumulated_chunks.append(chunk)

                    # vcprint(chunk, "CHUNK", color="cyan")

                    if chunk.candidates:
                        cand: Candidate
                        for cand in chunk.candidates:
                            if cand.content and cand.content.parts:
                                part: Part
                                for part in cand.content.parts:
                                    await self._handle_part(part, emitter)

                converted_response = self.translator.from_google(
                    accumulated_chunks, matrx_model_name
                )
            else:
                # Non-streaming mode - returns single GenerateContentResponse
                response: GenerateContentResponse = self.client.models.generate_content(
                    **config_data
                )

                # Wrap the single response in a list to maintain consistency with to_unified_config
                accumulated_chunks: list[GenerateContentResponse] = [response]

                # Send all content through emitter (same as streaming, but all at once)
                if response.candidates:
                    cand: Candidate
                    for cand in response.candidates:
                        if cand.content and cand.content.parts:
                            part: Part
                            for part in cand.content.parts:
                                await self._handle_part(part, emitter)

                converted_response = self.translator.from_google(
                    accumulated_chunks, matrx_model_name
                )

            return converted_response

        except Exception as e:
            vcprint(e, "[Google Chat] Error", color="red")
            # Import here to avoid circular dependency
            from matrx_ai.providers.errors import classify_google_error

            # Classify the error to determine if it's retryable
            error_info = classify_google_error(e)

            if emitter:
                await emitter.send_error(
                    error_type=error_info.error_type,
                    message=error_info.message,
                    user_message=error_info.user_message,
                )

            vcprint(e, "[Google Chat] Error", color="red")
            if "response" in locals():
                vcprint(response, "Response", color="red")
            traceback.print_exc()

            # Re-raise with error classification attached
            e.error_info = error_info
            raise

    async def _handle_part(self, part: Part, emitter: Emitter):
        await asyncio.sleep(0)

        # vcprint(part, "PART", color="green")

        if part.thought:
            if emitter:
                await emitter.send_chunk(
                    f"\n<reasoning>\n {part.text} \n</reasoning>\n"
                )
                # print(part.text)
                await asyncio.sleep(0)
            else:
                print(f"\n<reasoning>\n {part.text} \n</reasoning>\n")

        elif part.text:
            if emitter:
                await emitter.send_chunk(part.text)
                await asyncio.sleep(0)
            else:
                print(part.text)

        elif part.function_call:
            pass

        elif part.thought_signature:
            pass

        else:
            pass

        if LOCAL_DEBUG:
            await self._debug_part(part)

    async def _debug_part(self, part: Part):
        if part.thought:
            vcprint("=================== THOUGHT ===================", color="blue")
            print(part.text)
            vcprint("================================================", color="blue")
        elif part.text:
            print(part.text)
        elif part.function_call:
            vcprint(part.function_call, "Function Call", color="blue")
        elif part.thought_signature:
            vcprint("Empty Thought Signature", color="blue")
        else:
            vcprint(part, "Empty Part", color="blue")



"""
Comparison: Old approach vs New Agent system

This shows how the Agent system simplifies prompt handling.
"""

import asyncio

from matrx_utils import clear_terminal, vcprint

import matrx_ai
from matrx_ai.agents.definition import Agent
from matrx_ai.orchestrator.executor import execute_until_complete
from matrx_ai.orchestrator.requests import AIMatrixRequest
from matrx_ai.providers.unified_client import UnifiedAIClient
from tests.ai.test_context import create_test_execution_context

matrx_ai.initialize()
_ctx_token = create_test_execution_context()


async def test_autonomous_execution(settings_to_use):
    """Test autonomous execution that handles all tool calls automatically"""
    client = UnifiedAIClient()
    request = AIMatrixRequest.from_dict(settings_to_use)
    vcprint(request, "Request", color="green")

    try:
        completed = await execute_until_complete(
            initial_request=request,
            client=client,
            max_iterations=10,
            max_retries_per_iteration=0,
        )
        return completed
    except RuntimeError as e:
        vcprint(f"\n✗ Autonomous execution failed: {str(e)}", color="red")
        raise


# ============================================================================
# NEW APPROACH: Using Agent System (Clean & Simple)
# ============================================================================

async def create_conversation_with_agent(
    prompt_id: str,
    variables: dict,
    debug: bool = False,
):
    """Using Agent system - clean, type-safe, easy to understand"""
    agent = await Agent.from_prompt_id(prompt_id, variables=variables)
    
    conversation = await agent.execute(
        user_input="Please analyze this topic",
    )
    
    return conversation


# ============================================================================
# EVEN SIMPLER: One-liner approach
# ============================================================================

async def create_conversation_simple(
    prompt_id: str,
    variables: dict,
    debug: bool = False,
):
    """Simplest: Everything in one flow (conceptual example)"""
    agent = await Agent.from_prompt_id(prompt_id, variables=variables)
    return agent.to_config()


async def main():
    clear_terminal()
    
    print("=" * 70)
    print("AGENT SYSTEM COMPARISON")
    print("=" * 70)
    
    prompt_id = "002edf1b-26e6-4fb1-86cb-21000e319f2a"
    variables = {
        "topic": "Recent fires at a Swiss ski resort",
        "random": "IF YOU SEE THIS, DO NOT RESPOND",
    }
    
    print("\nUsing NEW Agent System:")
    print("-" * 70)
    
    print("\nBenefits:")
    print("  - Clean, type-safe API")
    print("  - Variables handled automatically")
    print("  - Config overrides support")
    print("  - Chainable methods")
    print("  - Easy to test and reuse")
    print("  - Self-documenting code")
    
    print("\n" + "=" * 70)
    print("Testing actual execution...")
    print("=" * 70)
    
    try:
        conversation = await create_conversation_with_agent(
            prompt_id, variables, debug=False
        )
        print("\nExecution completed successfully!")
        print(f"Conversation: {conversation}")
        print(f"Response preview: {conversation.last_response[:200]}...")
    except Exception as e:
        print(f"\nError (expected if model_id not configured): {e}")


if __name__ == "__main__":
    asyncio.run(main())



# File Location: /matrix/ai_models/ai_model_validator.py

import asyncio
import json
import re
from typing import Any

from matrx_utils import vcprint

from matrx_ai.db import _setup

_setup()

from matrx_ai.db.custom.ai_models.ai_model_manager import AiModelManager


class AiModelValidator(AiModelManager):
    """Validates and fixes the integrity of AI model data entries."""

    async def validate_data_integrity(self) -> dict[str, list[str]]:
        """
        Orchestrates the validation of all AI model entries and returns a report of issues.
        """
        models = await self.load_all_models()
        validation_report = {
            "name_issues": [],
            "model_class_issues": [],
            "endpoint_issues": [],
            "model_provider_issues": [],
            "max_tokens_issues": [],
            "capabilities_issues": [],
            "raw_jsonb_issues": [],
        }

        model_names = {model.name for model in models}

        for model in models:
            name_issues = self._validate_name(model)
            class_issues = self._validate_model_class(model, model_names)
            endpoint_issues = self._validate_endpoints(model)
            provider_issues = self._validate_model_provider(model)
            max_tokens_issues = self._validate_max_tokens(model)
            capabilities_issues = self._validate_capabilities(model)
            raw_jsonb_issues = self._validate_raw_jsonb_fields(model)

            model_identifier = f"Model '{model.name}' ({model.id})"
            if name_issues:
                validation_report["name_issues"].append(
                    f"{model_identifier}: {name_issues}"
                )
            if class_issues:
                validation_report["model_class_issues"].append(
                    f"{model_identifier}: {class_issues}"
                )
            if endpoint_issues:
                validation_report["endpoint_issues"].append(
                    f"{model_identifier}: {endpoint_issues}"
                )
                vcprint(
                    f"CRITICAL ENDPOINT ISSUE - {model_identifier}: {endpoint_issues}",
                    color="red",
                )
            if provider_issues:
                validation_report["model_provider_issues"].append(
                    f"{model_identifier}: {provider_issues}"
                )
            if max_tokens_issues:
                validation_report["max_tokens_issues"].append(
                    f"{model_identifier}: {max_tokens_issues}"
                )
            if capabilities_issues:
                validation_report["capabilities_issues"].append(
                    f"{model_identifier}: {capabilities_issues}"
                )
            if raw_jsonb_issues:
                validation_report["raw_jsonb_issues"].append(
                    f"{model_identifier}: {raw_jsonb_issues}"
                )
                vcprint(
                    f"RAW JSONB ISSUE - {model_identifier}: {raw_jsonb_issues}",
                    color="red",
                )

        self._print_validation_summary(validation_report)
        return validation_report

    async def fix_malformed_endpoints(self) -> dict[str, list[str]]:
        """
        Checks all models for malformed endpoints (strings that should be lists) and fixes them.
        Returns a report of fixes applied.
        """
        models = await self.load_all_models()
        fix_report = {"fixed": [], "skipped": [], "errors": []}

        for model in models:
            model_identifier = f"Model '{model.name}' ({model.id})"

            if not hasattr(model, "endpoints"):
                fix_report["skipped"].append(f"{model_identifier}: No endpoints field")
                continue

            endpoints = model.endpoints

            # If it's already a proper list of strings, skip
            if isinstance(endpoints, list) and all(
                isinstance(ep, str) for ep in endpoints
            ):
                fix_report["skipped"].append(
                    f"{model_identifier}: Endpoints already valid"
                )
                continue

            # If it's a string, try to fix it
            if isinstance(endpoints, str):
                try:
                    # Attempt to parse the string as JSON
                    parsed_endpoints = json.loads(endpoints)

                    # Ensure the parsed result is a list of strings
                    if not isinstance(parsed_endpoints, list):
                        fix_report["errors"].append(
                            f"{model_identifier}: Parsed endpoints is not a list"
                        )
                        continue

                    if not all(isinstance(ep, str) for ep in parsed_endpoints):
                        fix_report["errors"].append(
                            f"{model_identifier}: Parsed endpoints contains non-string elements"
                        )
                        continue

                    # Update the model with the fixed endpoints
                    await self.update_ai_model(model.id, endpoints=parsed_endpoints)
                    fix_report["fixed"].append(
                        f"{model_identifier}: Fixed endpoints from '{endpoints}' to {parsed_endpoints}"
                    )
                    vcprint(
                        f"FIXED - {model_identifier}: Converted endpoints to {parsed_endpoints}",
                        color="green",
                    )

                except json.JSONDecodeError:
                    fix_report["errors"].append(
                        f"{model_identifier}: Could not parse endpoints string '{endpoints}'"
                    )
                except Exception as e:
                    fix_report["errors"].append(
                        f"{model_identifier}: Error fixing endpoints - {str(e)}"
                    )
            else:
                fix_report["errors"].append(
                    f"{model_identifier}: Endpoints is neither a list nor a fixable string (type: {type(endpoints).__name__})"
                )

        self._print_fix_summary(fix_report, label="Endpoint")
        return fix_report

    def _validate_name(self, model: Any) -> str:
        """Validates that the model name exists and contains only allowed characters."""
        if not hasattr(model, "name") or not model.name:
            return "Missing or empty name"

        name = model.name
        if not re.match(r"^[a-zA-Z0-9-:./]+$", name):
            return f"Invalid characters in name '{name}' (allowed: alphanumeric, -, :, /, .; no spaces)"
        if " " in name:
            return f"Spaces not allowed in name '{name}'"
        return ""

    def _validate_model_class(self, model: Any, existing_names: set) -> str:
        """Validates that model_class exists and matches an existing model name pattern."""
        if not hasattr(model, "model_class") or not model.model_class:
            return "Missing or empty model_class"

        model_class = model.model_class
        is_valid_class = any(name.startswith(model_class) for name in existing_names)
        if not is_valid_class:
            return f"Model class '{model_class}' doesn't match any existing model name pattern"
        return ""

    def _validate_endpoints(self, model: Any) -> str:
        """Validates that endpoints is a non-empty list of strings."""
        if not hasattr(model, "endpoints"):
            return "Missing endpoints field"

        endpoints = model.endpoints
        if not isinstance(endpoints, list):
            return f"Endpoints must be a list, got {type(endpoints).__name__}"

        if not endpoints:
            return "Endpoints list is empty"

        if not all(isinstance(ep, str) for ep in endpoints):
            return "All endpoints must be strings"

        return ""

    def _validate_model_provider(self, model: Any) -> str:
        """Validates that model_provider exists and is not empty."""
        if not hasattr(model, "model_provider") or not model.model_provider:
            return "Missing or empty model_provider"
        return ""

    def _validate_max_tokens(self, model: Any) -> str:
        """Validates that max_tokens exists and is a positive integer."""
        if not hasattr(model, "max_tokens"):
            return "Missing max_tokens field"

        max_tokens = model.max_tokens
        if max_tokens is None:
            return "max_tokens is None"
        if not isinstance(max_tokens, int):
            return f"max_tokens must be an integer, got {type(max_tokens).__name__}"
        if max_tokens <= 0:
            return f"max_tokens must be positive, got {max_tokens}"
        return ""

    def _validate_capabilities(self, model: Any) -> str:
        """Validates that capabilities exists and is non-empty."""
        if not hasattr(model, "capabilities"):
            return "Missing capabilities field"

        capabilities = model.capabilities
        if capabilities is None:
            return "capabilities is None"
        if not isinstance(capabilities, (dict, list)):
            return f"capabilities must be a dict or list, got {type(capabilities).__name__}"
        if not capabilities:
            return "capabilities is empty"
        return ""

    def _validate_raw_jsonb_fields(self, model: Any) -> str:
        """
        Checks whether any JSONB field came back as a raw string instead of parsed JSON.
        This indicates the ORM failed to deserialize the value — typically caused by an
        empty string or malformed value stored in the DB column, which would crash older
        ORM versions with a JSONDecodeError.
        """
        jsonb_fields = ["endpoints", "capabilities", "controls"]
        raw_fields = []
        for field in jsonb_fields:
            value = getattr(model, field, None)
            if isinstance(value, str):
                raw_fields.append(f"{field}='{value[:60]}'" if len(value) > 60 else f"{field}='{value}'")
        if raw_fields:
            return f"JSONB fields returned as raw strings (unparsed): {', '.join(raw_fields)}"
        return ""

    async def fix_malformed_capabilities(self) -> dict[str, list[str]]:
        """
        Checks all models for malformed capabilities (strings that should be lists/dicts)
        and fixes them by JSON-parsing the raw string value.
        Returns a report of fixes applied.
        """
        models = await self.load_all_models()
        fix_report = {"fixed": [], "skipped": [], "errors": []}

        for model in models:
            model_identifier = f"Model '{model.name}' ({model.id})"

            if not hasattr(model, "capabilities"):
                fix_report["skipped"].append(f"{model_identifier}: No capabilities field")
                continue

            capabilities = model.capabilities

            # Already a valid non-empty list or dict — nothing to do
            if isinstance(capabilities, (list, dict)) and capabilities:
                fix_report["skipped"].append(
                    f"{model_identifier}: Capabilities already valid"
                )
                continue

            if isinstance(capabilities, str):
                try:
                    parsed = json.loads(capabilities)

                    if not isinstance(parsed, (list, dict)):
                        fix_report["errors"].append(
                            f"{model_identifier}: Parsed capabilities is not a list or dict"
                        )
                        continue

                    await self.update_ai_model(model.id, capabilities=parsed)
                    fix_report["fixed"].append(
                        f"{model_identifier}: Fixed capabilities — "
                        f"{type(parsed).__name__} with {len(parsed)} items: {parsed}"
                    )
                    vcprint(
                        f"FIXED - {model_identifier}: Converted capabilities to {parsed}",
                        color="green",
                    )

                except json.JSONDecodeError:
                    fix_report["errors"].append(
                        f"{model_identifier}: Could not parse capabilities string "
                        f"'{capabilities[:100]}'"
                    )
                except Exception as e:
                    fix_report["errors"].append(
                        f"{model_identifier}: Error fixing capabilities — {str(e)}"
                    )
            else:
                fix_report["errors"].append(
                    f"{model_identifier}: Capabilities is neither a valid structure nor a "
                    f"fixable string (type: {type(capabilities).__name__}, "
                    f"value: {str(capabilities)[:60]})"
                )

        self._print_fix_summary(fix_report, label="Capabilities")
        return fix_report

    def _print_validation_summary(self, report: dict[str, list[str]]) -> None:
        """Prints a summary of all validation issues found."""
        total_issues = sum(len(issues) for issues in report.values())
        if total_issues == 0:
            vcprint(
                "Data integrity validation completed: No issues found", color="green"
            )
            return

        vcprint(
            f"Data integrity validation completed: Found {total_issues} issues",
            color="yellow",
        )
        for category, issues in report.items():
            if issues:
                vcprint(f"\n{category.replace('_', ' ').title()}:", color="yellow")
                for issue in issues:
                    vcprint(f"- {issue}", color="yellow")

    def _print_fix_summary(self, report: dict[str, list[str]], label: str = "Endpoint") -> None:
        """Prints a summary of fixes applied for the given field label."""
        total_fixed = len(report["fixed"])
        total_skipped = len(report["skipped"])
        total_errors = len(report["errors"])

        vcprint(
            f"{label} fix process completed: {total_fixed} fixed, {total_skipped} skipped, {total_errors} errors",
            color="blue",
        )
        if total_fixed > 0:
            vcprint("\nFixed:", color="green")
            for fix in report["fixed"]:
                vcprint(f"- {fix}", color="green")
        if total_errors > 0:
            vcprint("\nErrors:", color="red")
            for error in report["errors"]:
                vcprint(f"- {error}", color="red")


async def validate_and_fix_endpoints():
    validator = AiModelValidator()
    # Run validation first
    validation_report = await validator.validate_data_integrity()
    vcprint(
        data=validation_report, title="Validation Report", pretty=True, color="blue"
    )

    # Fix malformed endpoints
    endpoint_fix_report = await validator.fix_malformed_endpoints()
    vcprint(data=endpoint_fix_report, title="Endpoint Fix Report", pretty=True, color="blue")

    # Fix malformed capabilities
    capabilities_fix_report = await validator.fix_malformed_capabilities()
    vcprint(data=capabilities_fix_report, title="Capabilities Fix Report", pretty=True, color="blue")


async def main():
    validator = AiModelValidator()
    validation_report = await validator.validate_data_integrity()
    vcprint(
        data=validation_report, title="Validation Report", pretty=True, color="blue"
    )


if __name__ == "__main__":
    from matrx_utils import clear_terminal
    clear_terminal()
    asyncio.run(validate_and_fix_endpoints())



from __future__ import annotations

import time
from collections import OrderedDict
from typing import TypeVar

T = TypeVar("T")

class TTLCache[T]:
    """Generic TTL + LRU cache.

    - Items expire after `ttl_seconds` of inactivity (last access, not creation).
    - When `max_size` is reached, the least-recently-used item is evicted.
    - `get()` refreshes the TTL (marks the item as recently used).
    - Thread-safe for single-threaded async (no locks needed with asyncio).
    """

    def __init__(self, ttl_seconds: int = 1800, max_size: int = 200):
        self._ttl = ttl_seconds
        self._max_size = max_size
        self._store: OrderedDict[str, T] = OrderedDict()
        self._timestamps: dict[str, float] = {}

    def get(self, key: str) -> T | None:
        if key not in self._store:
            return None

        if self._is_expired(key):
            self._remove(key)
            return None

        self._store.move_to_end(key)
        self._timestamps[key] = time.monotonic()
        return self._store[key]

    def set(self, key: str, value: T) -> None:
        if key in self._store:
            self._store.move_to_end(key)
            self._store[key] = value
            self._timestamps[key] = time.monotonic()
            return

        self._evict_expired()

        while len(self._store) >= self._max_size:
            oldest_key, _ = self._store.popitem(last=False)
            self._timestamps.pop(oldest_key, None)

        self._store[key] = value
        self._timestamps[key] = time.monotonic()

    def remove(self, key: str) -> None:
        self._remove(key)

    def exists(self, key: str) -> bool:
        if key not in self._store:
            return False
        if self._is_expired(key):
            self._remove(key)
            return False
        return True

    def clear(self) -> None:
        self._store.clear()
        self._timestamps.clear()

    @property
    def size(self) -> int:
        return len(self._store)

    def _is_expired(self, key: str) -> bool:
        ts = self._timestamps.get(key, 0)
        return (time.monotonic() - ts) > self._ttl

    def _remove(self, key: str) -> None:
        self._store.pop(key, None)
        self._timestamps.pop(key, None)

    def _evict_expired(self) -> None:
        now = time.monotonic()
        expired = [k for k, ts in self._timestamps.items() if (now - ts) > self._ttl]
        for k in expired:
            self._remove(k)



from enum import StrEnum
from typing import Any

from pydantic import BaseModel, Field


class AgentStatus(StrEnum):
    idle = "idle"
    running = "running"
    paused = "paused"
    completed = "completed"
    failed = "failed"


class AgentStartRequest(BaseModel):
    """Body for POST /api/ai/agents/{agent_id} — start a new agent conversation."""
    user_input: str | list[dict[str, Any]] | None = None
    variables: dict[str, Any] | None = None
    config_overrides: dict[str, Any] | None = None
    stream: bool = True
    debug: bool = False


class AgentInfo(BaseModel):
    agent_id: str
    status: AgentStatus
    created_at: str
    metadata: dict[str, Any] = Field(default_factory=dict)



# File: db/models.py
from matrx_orm import BigIntegerField, BooleanField, CharField, DateField, DateTimeField, DecimalField, EnumField, FloatField, ForeignKey, IntegerField, JSONBField, Model, SmallIntegerField, TextArrayField, TextField, UUIDField, model_registry, BaseDTO, BaseManager
from enum import Enum
from dataclasses import dataclass
from typing import ClassVar

class Users(Model):
    id = UUIDField(primary_key=True, null=False)
    email = CharField(null=False)

    _table_name = "users"
    _db_schema = "auth"
    _database = "supabase_automation_matrix"

class AiProvider(Model):
    id = UUIDField(primary_key=True, null=False)
    name = CharField()
    company_description = TextField()
    documentation_link = CharField()
    models_link = CharField()
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'ai_settings': {'from_model': 'AiSettings', 'from_field': 'ai_provider', 'referenced_field': 'id', 'related_name': 'ai_settings'}, 'ai_model': {'from_model': 'AiModel', 'from_field': 'model_provider', 'referenced_field': 'id', 'related_name': 'ai_model'}}
    _database = "supabase_automation_matrix"

class CxAgentMemory(Model):
    id = UUIDField(primary_key=True, null=False)
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False, unique=True)
    memory_type = TextField(null=False)
    scope = TextField(null=False, default='user', unique=True)
    scope_id = TextField(unique=True)
    key = TextField(null=False, unique=True)
    content = TextField(null=False)
    importance = FloatField(default=0.5)
    access_count = IntegerField(default=0)
    last_accessed_at = DateTimeField()
    expires_at = DateTimeField()
    metadata = JSONBField(null=False, default={})
    created_at = DateTimeField(null=False)
    updated_at = DateTimeField(null=False)
    deleted_at = DateTimeField()
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {}
    _database = "supabase_automation_matrix"

class Organizations(Model):
    id = UUIDField(primary_key=True, null=False)
    name = TextField(null=False)
    slug = TextField(null=False, unique=True)
    description = TextField()
    logo_url = TextField()
    website = TextField()
    created_at = DateTimeField()
    updated_at = DateTimeField()
    created_by = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    is_personal = BooleanField(default=False)
    settings = JSONBField(default={})
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'workspaces': {'from_model': 'Workspaces', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'workspaces'}, 'context_variables': {'from_model': 'ContextVariables', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'context_variables'}, 'permissions': {'from_model': 'Permissions', 'from_field': 'granted_to_organization_id', 'referenced_field': 'id', 'related_name': 'permissions'}, 'organization_members': {'from_model': 'OrganizationMembers', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'organization_members'}, 'user_active_context': {'from_model': 'UserActiveContext', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'user_active_context'}, 'broker_values': {'from_model': 'BrokerValues', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'broker_values'}, 'user_files': {'from_model': 'UserFiles', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'user_files'}, 'projects': {'from_model': 'Projects', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'projects'}, 'organization_invitations': {'from_model': 'OrganizationInvitations', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'organization_invitations'}, 'cx_conversation': {'from_model': 'CxConversation', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'cx_conversation'}, 'sandbox_instances': {'from_model': 'SandboxInstances', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'sandbox_instances'}, 'context_items': {'from_model': 'ContextItems', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'context_items'}, 'app_instances': {'from_model': 'AppInstances', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'app_instances'}, 'notes': {'from_model': 'Notes', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'notes'}, 'user_tables': {'from_model': 'UserTables', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'user_tables'}, 'transcripts': {'from_model': 'Transcripts', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'transcripts'}, 'workflow': {'from_model': 'Workflow', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'workflow'}, 'canvas_items': {'from_model': 'CanvasItems', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'canvas_items'}, 'prompts': {'from_model': 'Prompts', 'from_field': 'organization_id', 'referenced_field': 'id', 'related_name': 'prompts'}}
    _database = "supabase_automation_matrix"

class ShortcutCategories(Model):
    id = UUIDField(primary_key=True, null=False)
    placement_type = TextField(null=False)
    parent_category_id = ForeignKey(to_model='ShortcutCategories', to_column='id', )
    label = TextField(null=False)
    description = TextField()
    icon_name = TextField(null=False, default='SquareMenu')
    color = TextField(default='zinc')
    sort_order = IntegerField(default=999)
    is_active = BooleanField(default=True)
    metadata = JSONBField(default={})
    enabled_contexts = JSONBField(default=['general'])
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'content_blocks': {'from_model': 'ContentBlocks', 'from_field': 'category_id', 'referenced_field': 'id', 'related_name': 'content_blocks'}, 'prompt_shortcuts': {'from_model': 'PromptShortcuts', 'from_field': 'category_id', 'referenced_field': 'id', 'related_name': 'prompt_shortcuts'}, 'system_prompts_new': {'from_model': 'SystemPromptsNew', 'from_field': 'category_id', 'referenced_field': 'id', 'related_name': 'system_prompts_new'}}
    _database = "supabase_automation_matrix"

class Tools(Model):
    id = UUIDField(primary_key=True, null=False)
    name = TextField(null=False, unique=True)
    description = TextField(null=False)
    parameters = JSONBField(null=False)
    output_schema = JSONBField()
    annotations = JSONBField(default=[])
    function_path = TextField(null=False)
    category = TextField()
    tags = TextArrayField()
    icon = CharField(max_length=255)
    is_active = BooleanField(default=True)
    semver = TextField(default='1.0.0')
    created_at = DateTimeField()
    updated_at = DateTimeField()
    source_app = TextField(null=False, default='matrx_ai')
    version = IntegerField(null=False, default=1)
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'tool_versions': {'from_model': 'ToolVersions', 'from_field': 'tool_id', 'referenced_field': 'id', 'related_name': 'tool_versions'}, 'tool_test_samples': {'from_model': 'ToolTestSamples', 'from_field': 'tool_id', 'referenced_field': 'id', 'related_name': 'tool_test_samples'}, 'tool_ui_components': {'from_model': 'ToolUiComponents', 'from_field': 'tool_id', 'referenced_field': 'id', 'related_name': 'tool_ui_components'}}
    _database = "supabase_automation_matrix"

class AiModel(Model):
    id = UUIDField(primary_key=True, null=False)
    name = CharField(null=False)
    common_name = CharField()
    model_class = CharField(null=False)
    provider = CharField()
    endpoints = JSONBField()
    context_window = BigIntegerField()
    max_tokens = BigIntegerField()
    capabilities = JSONBField()
    controls = JSONBField()
    model_provider = ForeignKey(to_model=AiProvider, to_column='id', )
    is_deprecated = BooleanField(default=False)
    is_primary = BooleanField(default=False)
    is_premium = BooleanField(default=False)
    api_class = CharField()
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'ai_model_endpoint': {'from_model': 'AiModelEndpoint', 'from_field': 'ai_model_id', 'referenced_field': 'id', 'related_name': 'ai_model_endpoint'}, 'recipe_model': {'from_model': 'RecipeModel', 'from_field': 'ai_model', 'referenced_field': 'id', 'related_name': 'recipe_model'}, 'ai_settings': {'from_model': 'AiSettings', 'from_field': 'ai_model', 'referenced_field': 'id', 'related_name': 'ai_settings'}, 'agent_conversations': {'from_model': 'AgentConversations', 'from_field': 'model_id', 'referenced_field': 'id', 'related_name': 'agent_conversations'}, 'agent_requests': {'from_model': 'AgentRequests', 'from_field': 'model_id', 'referenced_field': 'id', 'related_name': 'agent_requests'}, 'cx_conversation': {'from_model': 'CxConversation', 'from_field': 'last_model_id', 'referenced_field': 'id', 'related_name': 'cx_conversation'}, 'cx_request': {'from_model': 'CxRequest', 'from_field': 'ai_model_id', 'referenced_field': 'id', 'related_name': 'cx_request'}, 'prompts': {'from_model': 'Prompts', 'from_field': 'model_id', 'referenced_field': 'id', 'related_name': 'prompts'}, 'prompt_builtins': {'from_model': 'PromptBuiltins', 'from_field': 'model_id', 'referenced_field': 'id', 'related_name': 'prompt_builtins'}}
    _database = "supabase_automation_matrix"

class ContentBlocks(Model):
    id = UUIDField(primary_key=True, null=False)
    block_id = CharField(null=False, max_length=100, unique=True)
    label = CharField(null=False, max_length=255)
    description = TextField()
    icon_name = CharField(null=False, max_length=100)
    template = TextField(null=False)
    sort_order = IntegerField(default=0)
    is_active = BooleanField(default=True)
    created_at = DateTimeField()
    updated_at = DateTimeField()
    category_id = ForeignKey(to_model=ShortcutCategories, to_column='id', )
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {}
    _database = "supabase_automation_matrix"

class Workspaces(Model):
    id = UUIDField(primary_key=True, null=False)
    organization_id = ForeignKey(to_model=Organizations, to_column='id', null=False)
    parent_workspace_id = ForeignKey(to_model='Workspaces', to_column='id', )
    name = TextField(null=False)
    description = TextField()
    settings = JSONBField(default={})
    created_at = DateTimeField()
    updated_at = DateTimeField()
    created_by = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'context_variables': {'from_model': 'ContextVariables', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'context_variables'}, 'user_active_context': {'from_model': 'UserActiveContext', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'user_active_context'}, 'broker_values': {'from_model': 'BrokerValues', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'broker_values'}, 'user_files': {'from_model': 'UserFiles', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'user_files'}, 'projects': {'from_model': 'Projects', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'projects'}, 'cx_conversation': {'from_model': 'CxConversation', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'cx_conversation'}, 'sandbox_instances': {'from_model': 'SandboxInstances', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'sandbox_instances'}, 'context_items': {'from_model': 'ContextItems', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'context_items'}, 'app_instances': {'from_model': 'AppInstances', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'app_instances'}, 'notes': {'from_model': 'Notes', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'notes'}, 'workspace_members': {'from_model': 'WorkspaceMembers', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'workspace_members'}, 'workspace_invitations': {'from_model': 'WorkspaceInvitations', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'workspace_invitations'}, 'user_tables': {'from_model': 'UserTables', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'user_tables'}, 'transcripts': {'from_model': 'Transcripts', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'transcripts'}, 'workflow': {'from_model': 'Workflow', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'workflow'}, 'canvas_items': {'from_model': 'CanvasItems', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'canvas_items'}, 'prompts': {'from_model': 'Prompts', 'from_field': 'workspace_id', 'referenced_field': 'id', 'related_name': 'prompts'}}
    _database = "supabase_automation_matrix"

class Projects(Model):
    id = UUIDField(primary_key=True, null=False)
    name = TextField(null=False)
    description = TextField()
    created_at = DateTimeField()
    updated_at = DateTimeField()
    created_by = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    organization_id = ForeignKey(to_model=Organizations, to_column='id', unique=True)
    slug = TextField(unique=True)
    settings = JSONBField(default={})
    is_personal = BooleanField(default=False)
    workspace_id = ForeignKey(to_model=Workspaces, to_column='id', )
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'context_variables': {'from_model': 'ContextVariables', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'context_variables'}, 'tasks': {'from_model': 'Tasks', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'tasks'}, 'user_active_context': {'from_model': 'UserActiveContext', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'user_active_context'}, 'broker_values': {'from_model': 'BrokerValues', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'broker_values'}, 'user_files': {'from_model': 'UserFiles', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'user_files'}, 'cx_conversation': {'from_model': 'CxConversation', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'cx_conversation'}, 'project_members': {'from_model': 'ProjectMembers', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'project_members'}, 'project_invitations': {'from_model': 'ProjectInvitations', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'project_invitations'}, 'sandbox_instances': {'from_model': 'SandboxInstances', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'sandbox_instances'}, 'context_items': {'from_model': 'ContextItems', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'context_items'}, 'app_instances': {'from_model': 'AppInstances', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'app_instances'}, 'notes': {'from_model': 'Notes', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'notes'}, 'rs_topic': {'from_model': 'RsTopic', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'rs_topic'}, 'user_tables': {'from_model': 'UserTables', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'user_tables'}, 'transcripts': {'from_model': 'Transcripts', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'transcripts'}, 'workflow': {'from_model': 'Workflow', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'workflow'}, 'canvas_items': {'from_model': 'CanvasItems', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'canvas_items'}, 'prompts': {'from_model': 'Prompts', 'from_field': 'project_id', 'referenced_field': 'id', 'related_name': 'prompts'}}
    _database = "supabase_automation_matrix"



class TaskPriority(str, Enum):
    HIGH = "high"
    LOW = "low"
    MEDIUM = "medium"

class Tasks(Model):
    id = UUIDField(primary_key=True, null=False)
    title = TextField(null=False)
    description = TextField()
    project_id = ForeignKey(to_model=Projects, to_column='id', )
    status = TextField(null=False, default='incomplete')
    due_date = DateField()
    created_at = DateTimeField()
    updated_at = DateTimeField()
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    authenticated_read = BooleanField(default=False)
    parent_task_id = ForeignKey(to_model='Tasks', to_column='id', )
    priority = EnumField(enum_class=TaskPriority, )
    assignee_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    settings = JSONBField(null=False, default={})
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'task_assignments': {'from_model': 'TaskAssignments', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'task_assignments'}, 'task_attachments': {'from_model': 'TaskAttachments', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'task_attachments'}, 'context_variables': {'from_model': 'ContextVariables', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'context_variables'}, 'task_comments': {'from_model': 'TaskComments', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'task_comments'}, 'user_active_context': {'from_model': 'UserActiveContext', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'user_active_context'}, 'broker_values': {'from_model': 'BrokerValues', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'broker_values'}, 'user_files': {'from_model': 'UserFiles', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'user_files'}, 'cx_conversation': {'from_model': 'CxConversation', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'cx_conversation'}, 'sandbox_instances': {'from_model': 'SandboxInstances', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'sandbox_instances'}, 'context_items': {'from_model': 'ContextItems', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'context_items'}, 'app_instances': {'from_model': 'AppInstances', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'app_instances'}, 'notes': {'from_model': 'Notes', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'notes'}, 'user_tables': {'from_model': 'UserTables', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'user_tables'}, 'transcripts': {'from_model': 'Transcripts', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'transcripts'}, 'workflow': {'from_model': 'Workflow', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'workflow'}, 'prompts': {'from_model': 'Prompts', 'from_field': 'task_id', 'referenced_field': 'id', 'related_name': 'prompts'}}
    _database = "supabase_automation_matrix"

class CxConversation(Model):
    id = UUIDField(primary_key=True, null=False)
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False)
    title = TextField()
    system_instruction = TextField()
    config = JSONBField(null=False, default={})
    status = TextField(null=False, default='active')
    message_count = SmallIntegerField(null=False, default=0)
    forked_from_id = ForeignKey(to_model='CxConversation', to_column='id', )
    forked_at_position = SmallIntegerField()
    created_at = DateTimeField(null=False)
    updated_at = DateTimeField(null=False)
    deleted_at = DateTimeField()
    metadata = JSONBField(null=False, default={})
    last_model_id = ForeignKey(to_model=AiModel, to_column='id', )
    parent_conversation_id = ForeignKey(to_model='CxConversation', to_column='id', )
    variables = JSONBField(null=False, default={})
    overrides = JSONBField(null=False, default={})
    description = TextField()
    keywords = TextArrayField()
    organization_id = ForeignKey(to_model=Organizations, to_column='id', )
    workspace_id = ForeignKey(to_model=Workspaces, to_column='id', )
    project_id = ForeignKey(to_model=Projects, to_column='id', )
    task_id = ForeignKey(to_model=Tasks, to_column='id', )
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'cx_tool_call': {'from_model': 'CxToolCall', 'from_field': 'conversation_id', 'referenced_field': 'id', 'related_name': 'cx_tool_call'}, 'cx_message': {'from_model': 'CxMessage', 'from_field': 'conversation_id', 'referenced_field': 'id', 'related_name': 'cx_message'}, 'cx_media': {'from_model': 'CxMedia', 'from_field': 'conversation_id', 'referenced_field': 'id', 'related_name': 'cx_media'}, 'cx_user_request': {'from_model': 'CxUserRequest', 'from_field': 'conversation_id', 'referenced_field': 'id', 'related_name': 'cx_user_request'}, 'cx_request': {'from_model': 'CxRequest', 'from_field': 'conversation_id', 'referenced_field': 'id', 'related_name': 'cx_request'}, 'canvas_items': {'from_model': 'CanvasItems', 'from_field': 'conversation_id', 'referenced_field': 'id', 'related_name': 'canvas_items'}}
    _database = "supabase_automation_matrix"

class Prompts(Model):
    id = UUIDField(primary_key=True, null=False)
    created_at = DateTimeField(null=False)
    updated_at = DateTimeField(null=False)
    name = CharField()
    messages = JSONBField()
    variable_defaults = JSONBField()
    tools = JSONBField()
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    settings = JSONBField()
    description = TextField()
    tags = TextArrayField(default=[])
    category = TextField()
    is_archived = BooleanField(null=False, default=False)
    is_favorite = BooleanField(null=False, default=False)
    model_id = ForeignKey(to_model=AiModel, to_column='id', )
    output_format = TextField()
    output_schema = JSONBField()
    dynamic_model = BooleanField(default=False)
    context_slots = JSONBField(null=False, default=[])
    organization_id = ForeignKey(to_model=Organizations, to_column='id', )
    workspace_id = ForeignKey(to_model=Workspaces, to_column='id', )
    project_id = ForeignKey(to_model=Projects, to_column='id', )
    task_id = ForeignKey(to_model=Tasks, to_column='id', )
    version = IntegerField(null=False, default=1)
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'system_prompts_new': {'from_model': 'SystemPromptsNew', 'from_field': 'source_prompt_id', 'referenced_field': 'id', 'related_name': 'system_prompts_new'}, 'prompt_actions': {'from_model': 'PromptActions', 'from_field': 'prompt_id', 'referenced_field': 'id', 'related_name': 'prompt_actions'}, 'system_prompts': {'from_model': 'SystemPrompts', 'from_field': 'source_prompt_id', 'referenced_field': 'id', 'related_name': 'system_prompts'}, 'prompt_versions': {'from_model': 'PromptVersions', 'from_field': 'prompt_id', 'referenced_field': 'id', 'related_name': 'prompt_versions'}, 'prompt_builtins': {'from_model': 'PromptBuiltins', 'from_field': 'source_prompt_id', 'referenced_field': 'id', 'related_name': 'prompt_builtins'}, 'prompt_apps': {'from_model': 'PromptApps', 'from_field': 'prompt_id', 'referenced_field': 'id', 'related_name': 'prompt_apps'}}
    _database = "supabase_automation_matrix"

class UserTables(Model):
    id = UUIDField(primary_key=True, null=False)
    table_name = CharField(null=False, max_length=255, unique=True)  # type: ignore[override]
    description = TextField()
    version = IntegerField(null=False, default=1)
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False, unique=True)
    is_public = BooleanField(null=False, default=False)
    authenticated_read = BooleanField(null=False, default=False)
    created_at = DateTimeField(null=False)
    updated_at = DateTimeField(null=False)
    row_ordering_config = JSONBField()
    organization_id = ForeignKey(to_model=Organizations, to_column='id', )
    workspace_id = ForeignKey(to_model=Workspaces, to_column='id', )
    project_id = ForeignKey(to_model=Projects, to_column='id', )
    task_id = ForeignKey(to_model=Tasks, to_column='id', )
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'table_data': {'from_model': 'TableData', 'from_field': 'table_id', 'referenced_field': 'id', 'related_name': 'table_data'}, 'table_fields': {'from_model': 'TableFields', 'from_field': 'table_id', 'referenced_field': 'id', 'related_name': 'table_fields'}}
    _database = "supabase_automation_matrix"

class CxMedia(Model):
    id = UUIDField(primary_key=True, null=False)
    conversation_id = ForeignKey(to_model=CxConversation, to_column='id', )
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False)
    kind = TextField(null=False)
    url = TextField(null=False)
    file_uri = TextField()
    mime_type = TextField()
    file_size_bytes = BigIntegerField()
    created_at = DateTimeField(null=False)
    deleted_at = DateTimeField()
    metadata = JSONBField(null=False, default={})
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {}
    _database = "supabase_automation_matrix"

class CxMessage(Model):
    id = UUIDField(primary_key=True, null=False)
    conversation_id = ForeignKey(to_model=CxConversation, to_column='id', null=False)
    role = TextField(null=False)
    position = SmallIntegerField(null=False)
    status = TextField(null=False, default='active')
    content = JSONBField(null=False, default=[])
    created_at = DateTimeField(null=False)
    deleted_at = DateTimeField()
    metadata = JSONBField(null=False, default={})
    content_history = JSONBField()
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'cx_tool_call': {'from_model': 'CxToolCall', 'from_field': 'message_id', 'referenced_field': 'id', 'related_name': 'cx_tool_call'}}
    _database = "supabase_automation_matrix"

class CxUserRequest(Model):
    id = UUIDField(primary_key=True, null=False)
    conversation_id = ForeignKey(to_model=CxConversation, to_column='id', null=False)
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False)
    trigger_message_position = SmallIntegerField()
    total_input_tokens = IntegerField(null=False, default=0)
    total_output_tokens = IntegerField(null=False, default=0)
    total_cached_tokens = IntegerField(null=False, default=0)
    total_tokens = IntegerField(null=False, default=0)
    total_cost = DecimalField(default=0)
    total_duration_ms = IntegerField()
    api_duration_ms = IntegerField()
    tool_duration_ms = IntegerField()
    iterations = SmallIntegerField(null=False, default=1)
    total_tool_calls = SmallIntegerField(null=False, default=0)
    status = TextField(null=False, default='pending')
    finish_reason = TextField()
    error = TextField()
    result_start_position = SmallIntegerField()
    result_end_position = SmallIntegerField()
    created_at = DateTimeField(null=False)
    completed_at = DateTimeField()
    deleted_at = DateTimeField()
    metadata = JSONBField(null=False, default={})
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'cx_tool_call': {'from_model': 'CxToolCall', 'from_field': 'user_request_id', 'referenced_field': 'id', 'related_name': 'cx_tool_call'}, 'cx_request': {'from_model': 'CxRequest', 'from_field': 'user_request_id', 'referenced_field': 'id', 'related_name': 'cx_request'}}
    _database = "supabase_automation_matrix"

class PromptBuiltins(Model):
    id = UUIDField(primary_key=True, null=False)
    created_at = DateTimeField(null=False)
    updated_at = DateTimeField(null=False)
    name = CharField(null=False)
    description = TextField()
    messages = JSONBField(null=False)
    variable_defaults = JSONBField()
    tools = JSONBField()
    settings = JSONBField()
    created_by_user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', )
    is_active = BooleanField(null=False, default=True)
    source_prompt_id = ForeignKey(to_model=Prompts, to_column='id', )
    source_prompt_snapshot_at = DateTimeField()
    tags = TextArrayField(default=[])
    category = TextField()
    model_id = ForeignKey(to_model=AiModel, to_column='id', )
    output_format = TextField()
    output_schema = JSONBField()
    is_archived = BooleanField(null=False, default=False)
    is_favorite = BooleanField(null=False, default=False)
    dynamic_model = BooleanField(default=False)
    context_slots = JSONBField(null=False, default=[])
    version = IntegerField(null=False, default=1)
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {'prompt_shortcuts': {'from_model': 'PromptShortcuts', 'from_field': 'prompt_builtin_id', 'referenced_field': 'id', 'related_name': 'prompt_shortcuts'}, 'prompt_actions': {'from_model': 'PromptActions', 'from_field': 'prompt_builtin_id', 'referenced_field': 'id', 'related_name': 'prompt_actions'}, 'prompt_builtin_versions': {'from_model': 'PromptBuiltinVersions', 'from_field': 'builtin_id', 'referenced_field': 'id', 'related_name': 'prompt_builtin_versions'}}
    _database = "supabase_automation_matrix"

class TableData(Model):
    id = UUIDField(primary_key=True, null=False, unique=True)
    table_id = ForeignKey(to_model=UserTables, to_column='id', null=False, unique=True)
    data = JSONBField(null=False)
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False)
    is_public = BooleanField(null=False, default=False)
    authenticated_read = BooleanField(null=False, default=False)
    created_at = DateTimeField(null=False)
    updated_at = DateTimeField(null=False)
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {}
    _database = "supabase_automation_matrix"

class CxRequest(Model):
    id = UUIDField(primary_key=True, null=False)
    user_request_id = ForeignKey(to_model=CxUserRequest, to_column='id', null=False)
    conversation_id = ForeignKey(to_model=CxConversation, to_column='id', null=False)
    api_class = TextField()
    iteration = SmallIntegerField(null=False, default=1)
    input_tokens = IntegerField()
    output_tokens = IntegerField()
    cached_tokens = IntegerField()
    total_tokens = IntegerField()
    cost = DecimalField()
    api_duration_ms = IntegerField()
    tool_duration_ms = IntegerField()
    total_duration_ms = IntegerField()
    tool_calls_count = SmallIntegerField(default=0)
    tool_calls_details = JSONBField()
    finish_reason = TextField()
    response_id = TextField()
    created_at = DateTimeField(null=False)
    deleted_at = DateTimeField()
    metadata = JSONBField(null=False, default={})
    ai_model_id = ForeignKey(to_model=AiModel, to_column='id', null=False)
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {}
    _database = "supabase_automation_matrix"

class CxToolCall(Model):
    id = UUIDField(primary_key=True, null=False)
    conversation_id = ForeignKey(to_model=CxConversation, to_column='id', null=False)
    message_id = ForeignKey(to_model=CxMessage, to_column='id', )
    user_id = ForeignKey(to_model=Users, to_column='id', to_schema='auth', null=False)
    user_request_id = ForeignKey(to_model=CxUserRequest, to_column='id', )
    tool_name = TextField(null=False)
    tool_type = TextField(null=False, default='local')
    call_id = TextField(null=False)
    status = TextField(null=False, default='pending')
    arguments = JSONBField(null=False, default={})
    success = BooleanField(null=False, default=True)
    output = TextField()
    output_type = TextField(default='text')
    is_error = BooleanField(default=False)
    error_type = TextField()
    error_message = TextField()
    duration_ms = IntegerField(null=False, default=0)
    started_at = DateTimeField(null=False)
    completed_at = DateTimeField(null=False)
    input_tokens = IntegerField(default=0)
    output_tokens = IntegerField(default=0)
    total_tokens = IntegerField(default=0)
    cost_usd = DecimalField(default=0)
    iteration = IntegerField(null=False, default=0)
    retry_count = IntegerField(default=0)
    parent_call_id = ForeignKey(to_model='CxToolCall', to_column='id', )
    execution_events = JSONBField(default=[])
    persist_key = TextField()
    file_path = TextField()
    metadata = JSONBField(null=False, default={})
    created_at = DateTimeField(null=False)
    deleted_at = DateTimeField()
    _inverse_foreign_keys: ClassVar[dict[str, dict[str, str]]] = {}
    _database = "supabase_automation_matrix"


model_registry.register_all(
[
        AiProvider,
        CxAgentMemory,
        Organizations,
        ShortcutCategories,
        Tools,
        AiModel,
        ContentBlocks,
        Workspaces,
        Projects,
        Tasks,
        CxConversation,
        Prompts,
        UserTables,
        CxMedia,
        CxMessage,
        CxUserRequest,
        PromptBuiltins,
        TableData,
        CxRequest,
        CxToolCall,
        Users
    ]
)



@dataclass
class AiProviderDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: AiProvider):
        return cls(id=str(model.id))


class AiProviderManager(BaseManager):
    def __init__(self):
        super().__init__(AiProvider, AiProviderDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: AiProvider) -> None:
        pass
    


@dataclass
class CxAgentMemoryDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxAgentMemory):
        return cls(id=str(model.id))


class CxAgentMemoryManager(BaseManager):
    def __init__(self):
        super().__init__(CxAgentMemory, CxAgentMemoryDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxAgentMemory) -> None:
        pass
    


@dataclass
class OrganizationsDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: Organizations):
        return cls(id=str(model.id))


class OrganizationsManager(BaseManager):
    def __init__(self):
        super().__init__(Organizations, OrganizationsDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Organizations) -> None:
        pass
    


@dataclass
class ShortcutCategoriesDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: ShortcutCategories):
        return cls(id=str(model.id))


class ShortcutCategoriesManager(BaseManager):
    def __init__(self):
        super().__init__(ShortcutCategories, ShortcutCategoriesDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: ShortcutCategories) -> None:
        pass
    


@dataclass
class ToolsDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: Tools):
        return cls(id=str(model.id))


class ToolsManager(BaseManager):
    def __init__(self):
        super().__init__(Tools, ToolsDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Tools) -> None:
        pass
    


@dataclass
class AiModelDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: AiModel):
        return cls(id=str(model.id))


class AiModelManager(BaseManager):
    def __init__(self):
        super().__init__(AiModel, AiModelDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: AiModel) -> None:
        pass
    


@dataclass
class ContentBlocksDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: ContentBlocks):
        return cls(id=str(model.id))


class ContentBlocksManager(BaseManager):
    def __init__(self):
        super().__init__(ContentBlocks, ContentBlocksDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: ContentBlocks) -> None:
        pass
    


@dataclass
class WorkspacesDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: Workspaces):
        return cls(id=str(model.id))


class WorkspacesManager(BaseManager):
    def __init__(self):
        super().__init__(Workspaces, WorkspacesDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Workspaces) -> None:
        pass
    


@dataclass
class ProjectsDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: Projects):
        return cls(id=str(model.id))


class ProjectsManager(BaseManager):
    def __init__(self):
        super().__init__(Projects, ProjectsDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Projects) -> None:
        pass
    


@dataclass
class TasksDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: Tasks):
        return cls(id=str(model.id))


class TasksManager(BaseManager):
    def __init__(self):
        super().__init__(Tasks, TasksDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Tasks) -> None:
        pass
    


@dataclass
class CxConversationDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxConversation):
        return cls(id=str(model.id))


class CxConversationManager(BaseManager):
    def __init__(self):
        super().__init__(CxConversation, CxConversationDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxConversation) -> None:
        pass
    


@dataclass
class PromptsDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: Prompts):
        return cls(id=str(model.id))


class PromptsManager(BaseManager):
    def __init__(self):
        super().__init__(Prompts, PromptsDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: Prompts) -> None:
        pass
    


@dataclass
class UserTablesDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: UserTables):
        return cls(id=str(model.id))


class UserTablesManager(BaseManager):
    def __init__(self):
        super().__init__(UserTables, UserTablesDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: UserTables) -> None:
        pass
    


@dataclass
class CxMediaDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxMedia):
        return cls(id=str(model.id))


class CxMediaManager(BaseManager):
    def __init__(self):
        super().__init__(CxMedia, CxMediaDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxMedia) -> None:
        pass
    


@dataclass
class CxMessageDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxMessage):
        return cls(id=str(model.id))


class CxMessageManager(BaseManager):
    def __init__(self):
        super().__init__(CxMessage, CxMessageDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxMessage) -> None:
        pass
    


@dataclass
class CxUserRequestDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxUserRequest):
        return cls(id=str(model.id))


class CxUserRequestManager(BaseManager):
    def __init__(self):
        super().__init__(CxUserRequest, CxUserRequestDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxUserRequest) -> None:
        pass
    


@dataclass
class PromptBuiltinsDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: PromptBuiltins):
        return cls(id=str(model.id))


class PromptBuiltinsManager(BaseManager):
    def __init__(self):
        super().__init__(PromptBuiltins, PromptBuiltinsDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: PromptBuiltins) -> None:
        pass
    


@dataclass
class TableDataDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: TableData):
        return cls(id=str(model.id))


class TableDataManager(BaseManager):
    def __init__(self):
        super().__init__(TableData, TableDataDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: TableData) -> None:
        pass
    


@dataclass
class CxRequestDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxRequest):
        return cls(id=str(model.id))


class CxRequestManager(BaseManager):
    def __init__(self):
        super().__init__(CxRequest, CxRequestDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxRequest) -> None:
        pass
    


@dataclass
class CxToolCallDTO(BaseDTO):
    id: str

    @classmethod
    async def from_model(cls, model: CxToolCall):
        return cls(id=str(model.id))


class CxToolCallManager(BaseManager):
    def __init__(self):
        super().__init__(CxToolCall, CxToolCallDTO)

    def _initialize_manager(self):
        super()._initialize_manager()

    async def _initialize_runtime_data(self, item: CxToolCall) -> None:
        pass